#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import re
import tarfile
import time
import uuid
from collections import OrderedDict
from typing import List

import yaml

from kubemarine.core import utils, flow
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.resources import DynamicResources
from kubemarine.procedures import install, backup
from kubemarine import system, kubernetes, etcd


def missing_or_empty(file: str) -> bool:
    if not os.path.exists(file):
        return True
    content = utils.read_external(file)
    if re.search(r'^\s*$', content):
        return True

    return False


def replace_config_from_backup_if_needed(procedure_inventory_filepath: str, config: str) -> None:
    # pylint: disable=bad-builtin

    if missing_or_empty(config):
        print('Config is missing or empty - retrieving config from backup archive...')
        procedure = utils.load_yaml(procedure_inventory_filepath)
        if not procedure:
            procedure = {}
        backup_location = procedure.get("backup_location")
        if not backup_location:
            raise Exception('Backup location is not specified in procedure')

        print('Unpacking cluster.yaml...')
        with tarfile.open(backup_location, 'r:gz') as tar:
            member = tar.getmember('original_cluster.yaml')
            tar.makefile(member, config)
            tar.close()


def unpack_data(resources: DynamicResources) -> None:
    logger = resources.logger()
    context = resources.context
    backup_tmp_directory = backup.prepare_backup_tmpdir(logger, context)
    backup_file_source = resources.procedure_inventory().get('backup_location')

    if not backup_file_source:
        raise Exception('Backup source not specified in procedure')

    backup_file_source = utils.get_external_resource_path(backup_file_source)
    if not os.path.isfile(backup_file_source):
        raise FileNotFoundError('Backup file "%s" not found' % backup_file_source)

    logger.debug('Unpacking all data...')
    with tarfile.open(backup_file_source, 'r:gz') as tar:
        for member in tar:
            if member.isdir():
                continue
            fname = os.path.join(backup_tmp_directory, member.name)
            logger.debug(fname)
            fname_parts = fname.split('/')
            if len(fname_parts) > 1:
                fname_dir = "/".join(fname_parts[:-1])
                if not os.path.isdir(fname_dir):
                    os.makedirs(fname_dir, exist_ok=True)
            tar.makefile(member, fname)
        tar.close()

    descriptor_filepath = os.path.join(backup_tmp_directory, 'descriptor.yaml')
    if not os.path.isfile(descriptor_filepath):
        raise FileNotFoundError('Descriptor not found in backup file')

    with utils.open_external(descriptor_filepath, 'r') as stream:
        context['backup_descriptor'] = yaml.safe_load(stream)


def stop_cluster(cluster: KubernetesCluster) -> None:
    cluster.log.debug('Stopping the existing cluster...')
    result = cluster.nodes['control-plane'].sudo(
        'systemctl stop kubelet; '
        'sudo crictl rm -fa; '
        'sudo crictl ps -a; '
        'sudo rm -rf /var/lib/etcd; '
        'sudo mkdir -p /var/lib/etcd',
        pty=True, warn=True)
    cluster.log.verbose(result)


def import_nodes_data(cluster: KubernetesCluster) -> None:
    with cluster.nodes['all'].new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            node_name = node.get_node_name()
            cluster.log.debug('Uploading backup for \'%s\'' % node_name)
            node.put(os.path.join(cluster.context['backup_tmpdir'], 'nodes_data', '%s.tar.gz' % node_name),
                     '/tmp/kubemarine-backup.tar.gz', compare_hashes=True)


def restore_dns_resolv_conf(cluster: KubernetesCluster) -> None:
    import_nodes_data(cluster)

    unpack_cmd = "sudo tar xzvf /tmp/kubemarine-backup.tar.gz -C / --overwrite /etc/resolv.conf"
    result = cluster.nodes['all'].sudo(
        f"readlink /etc/resolv.conf ; "
        f"if [ $? -ne 0 ]; then sudo chattr -i /etc/resolv.conf; {unpack_cmd} && sudo chattr +i /etc/resolv.conf; "
        f"fi ", pty=True)

    cluster.log.debug(result)


def restore_thirdparties(cluster: KubernetesCluster) -> None:
    install.system_prepare_thirdparties(cluster)


def import_nodes(cluster: KubernetesCluster) -> None:
    if not cluster.is_task_completed('restore.dns.resolv_conf'):
        import_nodes_data(cluster)

    cluster.log.debug('Unpacking backup...')

    unpack_cmd = "sudo tar xzvf /tmp/kubemarine-backup.tar.gz -C / --overwrite --exclude /etc/resolv.conf"
    result = cluster.nodes['all'].sudo(unpack_cmd, pty=True)

    cluster.log.debug(result)


def import_etcd(cluster: KubernetesCluster) -> None:
    etcd_all_certificates = cluster.procedure_inventory.get('restore_plan', {}).get('etcd', {}).get('certificates', {})
    etcd_cert = etcd_all_certificates.get('cert', cluster.globals['etcd']['default_arguments']['cert'])
    etcd_key = etcd_all_certificates.get('key', cluster.globals['etcd']['default_arguments']['key'])
    etcd_cacert = etcd_all_certificates.get('cacert', cluster.globals['etcd']['default_arguments']['cacert'])
    etcd_peer_cert = etcd_all_certificates.get('peer_cert', cluster.globals['etcd']['default_arguments']['peer_cert'])
    etcd_peer_key = etcd_all_certificates.get('peer_key', cluster.globals['etcd']['default_arguments']['peer_key'])
    etcd_peer_cacert = etcd_all_certificates.get('peer_cacert',
                                                 cluster.globals['etcd']['default_arguments']['peer_cacert'])

    etcd_image = cluster.procedure_inventory.get('restore_plan', {}).get('etcd', {}).get('image')
    if not etcd_image:
        etcd_image = cluster.context['backup_descriptor'].get('etcd', {}).get('image')
    if not etcd_image:
        raise Exception('Unknown ETCD image to restore from')
    cluster.log.verbose('ETCD will be restored from the following image: ' + etcd_image)

    cluster.log.debug('Uploading ETCD snapshot...')
    snap_name = '/var/lib/etcd/etcd-snapshot%s.db' % int(round(time.time() * 1000))
    cluster.nodes['control-plane'].put(os.path.join(cluster.context['backup_tmpdir'], 'etcd.db'), snap_name,
                                       sudo=True, compare_hashes=True)

    initial_cluster_list = []
    initial_cluster_list_without_names = []
    for control_plane in cluster.nodes['control-plane'].get_ordered_members_configs_list():
        initial_cluster_list.append(control_plane['name'] + '=https://' + control_plane["internal_address"] + ":2380")
        initial_cluster_list_without_names.append('https://' + control_plane["internal_address"] + ":2379")
    initial_cluster = ','.join(initial_cluster_list)

    container_name = f'etcd-{uuid.uuid4().hex}'
    mount_options = '-mount type=bind,src=/var/lib/etcd,dst=/var/lib/etcd,options=rbind:rw ' \
                    '-mount type=bind,src=/etc/kubernetes/pki/etcd,dst=/etc/kubernetes/pki/etcd,options=rbind:rw'

    etcd_instances = 0
    for control_plane in cluster.nodes['control-plane'].get_ordered_members_configs_list():
        cluster.log.debug('Restoring ETCD member ' + control_plane['name'])
        control_plane_conn = cluster.make_group([control_plane['connect_to']])
        control_plane_conn.sudo(
            f'chmod 777 {snap_name} && '
            f'sudo ls -la {snap_name} && '
            f'sudo ETCD_IMAGE="{etcd_image}" ETCD_MOUNTS="{mount_options}" etcdctl '
            f'--cert={etcd_cert} '
            f'--key={etcd_key} '
            f'--cacert={etcd_cacert} '
            f'--endpoints={",".join(initial_cluster_list_without_names)} '
            f'snapshot restore {snap_name} '
            f'--name={control_plane["name"]} '
            f'--data-dir=/var/lib/etcd/snapshot '
            f'--initial-cluster={initial_cluster} '
            f'--initial-advertise-peer-urls=https://{control_plane["internal_address"]}:2380',
            hide=False, pty=True)

        _ = control_plane_conn.sudo(
            f'mv /var/lib/etcd/snapshot/member /var/lib/etcd/member && '
            f'sudo rm -rf /var/lib/etcd/snapshot {snap_name} && '
            f'sudo ctr run -d --net-host '
            f'--env ETCDCTL_API=3 '
            f'{mount_options} '
            f'{etcd_image} {container_name} etcd '
            f'--advertise-client-urls=https://{control_plane["internal_address"]}:2379 '
            f'--cert-file={etcd_cert} '
            f'--key-file={etcd_key} '
            f'--trusted-ca-file={etcd_cacert} '
            f'--client-cert-auth=true '
            f'--data-dir=/var/lib/etcd '
            f'--initial-advertise-peer-urls=https://{control_plane["internal_address"]}:2380 '
            f'--initial-cluster={initial_cluster} '
            f'--listen-client-urls=https://127.0.0.1:2379,https://{control_plane["internal_address"]}:2379 '
            f'--listen-peer-urls=https://{control_plane["internal_address"]}:2380 '
            f'--name={control_plane["name"]} '
            f'--peer-client-cert-auth=true '
            f'--peer-cert-file={etcd_peer_cert} '
            f'--peer-key-file={etcd_peer_key} '
            f'--peer-trusted-ca-file={etcd_peer_cacert} ',
            pty=True)

        etcd_instances += 1

    # After restore check db size equal, cluster health and leader elected
    # Checks should be changed
    cluster_status = etcd.wait_for_health(cluster, cluster.nodes['control-plane'].get_any_member())

    # Check DB size is correct
    backup_source = cluster.context['backup_descriptor'].get('etcd', {}).get('source')
    backup_descriptor_dbsize = None
    if backup_source:
        backup_descriptor_dbsize = cluster.context['backup_descriptor'].get('etcd', {}).get('status', {}) \
            .get(backup_source, {}).get('status', {}).get('dbsize')
    if backup_descriptor_dbsize:
        expected_dbsize = int(backup_descriptor_dbsize)
        for item in cluster_status:
            real_dbsize = int(item.get('status', {}).get('dbsize'))
            if not real_dbsize:
                raise Exception('ETCD member "%s" do not have DB size' % item.get('endpoint'))
            cluster.log.verbose('Endpoint "%s" DB real size %s, expected size %s'
                                % (item.get('endpoint'), expected_dbsize, real_dbsize))
            # restored db should have equal or greater DB size
            if expected_dbsize > real_dbsize:
                raise Exception('ETCD member "%s" has invalid DB size' % item.get('endpoint'))
        cluster.log.debug('DB size "%s" is correct' % expected_dbsize)
    else:
        cluster.log.verbose('It is not possible to verify db size - descriptor do not contain such information')

    # Stop and remove container
    cluster.nodes['control-plane'].sudo(
        f"ctr task rm -f {container_name} && "
        f"sudo ctr container rm {container_name}",
        pty=True)


def reboot(cluster: KubernetesCluster) -> None:
    system.reboot_group(cluster.nodes['all'], try_graceful=False)
    kubernetes.wait_for_nodes(cluster.nodes['control-plane'])


tasks = OrderedDict({
    "prepare": {
        "stop_cluster": stop_cluster,
    },
    "restore": {
        "dns": {
            "resolv_conf": restore_dns_resolv_conf,
        },
        "thirdparties": restore_thirdparties,
    },
    "import": {
        "nodes": import_nodes,
        "etcd": import_etcd
    },
    "reboot": reboot
})


class RestoreFlow(flow.Flow):
    def _run(self, resources: DynamicResources) -> None:
        unpack_data(resources)
        flow.run_actions(resources, [RestoreAction()])


class RestoreAction(flow.TasksAction):
    def __init__(self) -> None:
        super().__init__('restore', tasks, recreate_inventory=True)


def create_context(cli_arguments: List[str] = None) -> dict:
    cli_help = '''
    Script for restoring Kubernetes resources and nodes contents from backup file.

    How to use:

    '''

    parser = flow.new_procedure_parser(cli_help, tasks=tasks)

    context = flow.create_context(parser, cli_arguments, procedure='restore')
    context['backup_descriptor'] = {}

    return context


def main(cli_arguments: List[str] = None) -> None:
    context = create_context(cli_arguments)
    args = context['execution_arguments']

    replace_config_from_backup_if_needed(args['procedure_config'], args['config'])

    RestoreFlow().run_flow(context)


if __name__ == '__main__':
    main()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import time
from typing import List, Optional

from jinja2 import Template

from kubemarine import system, packages
from kubemarine.core import utils, static
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.core.group import (
    NodeGroup, RunnersGroupResult, NodeConfig, DeferredGroup, RunResult, AbstractGroup, CollectorCallback
)

ERROR_VRRP_IS_NOT_CONFIGURED = \
    'Balancer is combined with other role, but there is no VRRP IP configured for node \'%s\'.'

ERROR_NO_BOUND_VRRP_CONFIGURED_MNTC = \
    'No suitable bindings found for haproxy in maintenance mode for node \'%s\'. ' \
    'Balancer is combined with other role and has no configured VRRP IP ' \
    'that is not marked with maintenance-type: "not bind"'


def is_maintenance_mode(cluster: KubernetesCluster) -> bool:
    maintenance_mode: bool = cluster.inventory['services']['loadbalancer']['haproxy']['maintenance_mode']
    return maintenance_mode


def _get_associations_for_node(node: AbstractGroup[RunResult]) -> dict:
    return node.cluster.get_associations_for_node(node.get_host(), 'haproxy')


def is_vrrp_not_bind(vrrp_item: dict) -> bool:
    maintenance_type: str = vrrp_item.get('params', {}).get('maintenance-type', '')
    return maintenance_type == 'not bind'


def _get_bindings(inventory: dict, node: NodeConfig, *, maintenance: bool) -> List[str]:
    # bindings list for common config and maintenance should be different

    is_combined = len(node['roles']) > 1

    if not maintenance and not is_combined:
        return ["0.0.0.0", '::']

    # If we have combination of balancer-control-plane / balancer-worker or if haproxy is in maintenance mode,
    # VRRP IP should be defined for the given balancer.
    # In maintenance mode it should also be not "not bind".

    bindings = []
    for item in inventory['vrrp_ips']:
        # skip IPs with type 'not bind' in maintenance mode
        if maintenance and is_vrrp_not_bind(item):
            continue
        for record in item['hosts']:
            if record['name'] == node['name']:
                bindings.append(item['ip'])

    if not is_combined:
        # In maintenance mode and if balancer is not combined with some other role,
        # we can listen also own internal address of the balancer
        bindings.append(node['internal_address'])

    # remove duplicates
    return list(set(bindings))


@enrichment(EnrichmentStage.FULL)
def enrich_inventory(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory

    for group in cluster.make_group_from_roles(['balancer']).get_accessible_nodes().get_ordered_members_list():
        node = group.get_config()

        regular_bindings = _get_bindings(inventory, node, maintenance=False)
        if not regular_bindings:
            raise Exception(ERROR_VRRP_IS_NOT_CONFIGURED % node['name'])

        is_mntc_mode = is_maintenance_mode(cluster)
        mntc_bindings = _get_bindings(inventory, node, maintenance=True)
        if is_mntc_mode and not mntc_bindings:
            raise Exception(ERROR_NO_BOUND_VRRP_CONFIGURED_MNTC % node["name"])

        not_bind = sum(1 for item in inventory["vrrp_ips"] if is_vrrp_not_bind(item))
        if bool(not_bind) != is_mntc_mode:
            raise Exception("Haproxy maintenance mode should be used when and only when "
                            "there is at least one VRRP IP with 'maintenance-type: not bind'")

        os_family = group.get_nodes_os()
        if is_mntc_mode and os_family not in ('unknown', 'unsupported'):
            config_location = _get_associations_for_node(group)['config_location']
            mntc_config_location = inventory['services']['loadbalancer']['haproxy']['mntc_config_location']
            # if 'maintenance_mode' is True then maintenance config and default config must be stored in different files'
            if mntc_config_location == config_location:
                raise Exception("Maintenance mode configuration file must be different with default configuration file")


def get_config_path(group: NodeGroup) -> RunnersGroupResult:
    collector = CollectorCallback(group.cluster)
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            package_associations = _get_associations_for_node(node)
            cmd = f"systemctl show -p MainPID {package_associations['service_name']} " \
                  f"| cut -d '=' -f2 " \
                  f"| xargs -I PID sudo cat /proc/PID/environ " \
                  f"| tr '\\0' '\\n' | grep CONFIG | cut -d \"=\" -f2 | tr -d '\\n'"
            node.sudo(cmd, callback=collector)

    return collector.result


def install(group: NodeGroup) -> RunnersGroupResult:
    cluster = group.cluster
    defer = group.new_defer()
    collector = CollectorCallback(cluster)
    for node in defer.get_ordered_members_list():
        package_associations = _get_associations_for_node(node)
        node.sudo("%s -v" % package_associations['executable_name'], warn=True, callback=collector)

    defer.flush()

    if not collector.result.is_any_failed():
        cluster.log.debug("HAProxy already installed, nothing to install")
    else:
        collector = CollectorCallback(cluster)
        for node in defer.get_ordered_members_list():
            package_associations = _get_associations_for_node(node)
            packages.install(node, include=package_associations['package_name'], pty=True, callback=collector)

        defer.flush()

    for node in defer.get_ordered_members_list():
        package_associations = _get_associations_for_node(node)
        service_name = package_associations['service_name']
        patch_path = "./resources/drop_ins/haproxy.conf"
        node.call(system.patch_systemd_service, service_name=service_name, patch_source=patch_path)
        enable(node)

    defer.flush()
    return collector.result


def restart(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    cluster.log.debug("Restarting haproxy in all group...")
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            service_name = _get_associations_for_node(node)['service_name']
            system.restart_service(node, name=service_name)

    cluster.log.debug("Sleep while haproxy comes-up...")
    time.sleep(static.GLOBALS['haproxy']['restart_wait'])


def disable(group: NodeGroup) -> None:
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            service_name = _get_associations_for_node(node)['service_name']
            system.disable_service(node, name=service_name)


def enable(node: DeferredGroup) -> None:
    # currently it is invoked only for single node
    service_name = _get_associations_for_node(node)['service_name']
    system.enable_service(node, name=service_name, now=True)


def get_config(cluster: KubernetesCluster, node: NodeConfig, maintenance: bool = False) -> str:
    future_nodes = cluster.nodes['all'].get_ordered_members_configs_list()

    inventory = cluster.inventory
    bindings = _get_bindings(inventory, node, maintenance=maintenance)

    config_options: dict = inventory['services']['loadbalancer']['haproxy']
    config_string: Optional[str] = config_options.get('config')
    if config_string is not None:
        return config_string

    target_ports: dict = inventory['services']['loadbalancer']['target_ports']

    # todo support custom template for maintenance mode
    if not maintenance and config_options.get('config_file'):
        config_source = utils.read_external(config_options['config_file'])
    else:
        config_source = utils.read_internal('templates/haproxy.cfg.j2')

    return Template(config_source).render(nodes=future_nodes,
                                          bindings=bindings,
                                          config_options=config_options,
                                          target_ports=target_ports)


def configure(group: DeferredGroup) -> None:
    cluster: KubernetesCluster = group.cluster

    for node in group.get_ordered_members_list():
        node_config = node.get_config()
        node_name = node.get_node_name()
        package_associations = _get_associations_for_node(node)

        cluster.log.debug("\nConfiguring haproxy on \'%s\'..." % node_name)
        config = get_config(cluster, node_config)
        utils.dump_file(cluster, config, 'haproxy_%s.cfg' % node_name)
        node.put(io.StringIO(config), package_associations['config_location'], backup=True, sudo=True, mkdir=True)

        # add maintenance config to balancer if 'maintenance_mode' is True
        if is_maintenance_mode(cluster):
            mntc_config_location = cluster.inventory['services']['loadbalancer']['haproxy']['mntc_config_location']
            cluster.log.debug("\nConfiguring haproxy for maintenance on \'%s\'..." % node_name)
            mntc_config = get_config(cluster, node_config, True)
            utils.dump_file(cluster, mntc_config, 'haproxy_mntc_%s.cfg' % node_name)
            node.put(io.StringIO(mntc_config), mntc_config_location, backup=True, sudo=True, mkdir=True)


def override_haproxy18(group: DeferredGroup) -> None:
    rhel_nodes = group.get_subgroup_with_os('rhel')
    if rhel_nodes.is_empty():
        group.cluster.log.debug('Haproxy18 override is not required')
        return

    # Any node in group has rhel OS family, so association can be fetched from any node.
    any_host = rhel_nodes.get_first_member().get_host()
    package_associations = group.cluster.get_associations_for_node(any_host, 'haproxy')
    # TODO: Do not replace the whole file, replace only parameter
    rhel_nodes.put(
        io.StringIO("CONFIG=%s\n" % package_associations['config_location']),
        '/etc/sysconfig/%s' % package_associations['service_name'],
        backup=True, sudo=True)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import textwrap
from traceback import print_exc
import csv
from datetime import datetime
from types import TracebackType
from typing import Dict, Optional, Type, Union, List

from kubemarine.core import utils, log

from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.executor import BaseExecutorException

TC_UNKNOWN = -1
TC_PASSED = 0
TC_FAILED = 1
TC_WARNED = 2
TC_EXCEPTED = 3

badges_weights = {
    'succeeded': 0,
    'warned': 1,
    'failed': 2,
    'excepted': 3,
    'unknown': 4,
}


class TestCase:

    def __enter__(self) -> 'TestCase':
        return self

    def __exit__(self, _: Optional[Type[BaseException]], value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> bool:
        if value is None:
            if self.status is TC_UNKNOWN:
                self.success()
        elif isinstance(value, TestFailure):
            self.fail(value)
        elif isinstance(value, TestWarn):
            self.warn(value)
        else:
            self.exception(value)
        print(self.get_summary(show_hint=True))  # pylint: disable=bad-builtin
        if isinstance(value, KeyboardInterrupt):
            # No need to print output of remote commands twice.
            raise KeyboardInterrupt()

        return True

    def __init__(self, cluster: KubernetesCluster, id_: str, category: str, name: str,
                 default_results: str = None, minimal: int = None, recommended: int = None):
        self.include_in_ts(cluster.context['testsuite'])
        self.category = category
        self.id = str(id_)
        self.name = name
        self.status = TC_UNKNOWN
        self.results: Union[str, BaseException, None] = default_results
        self.minimal = minimal
        self.recommended = recommended
        self.cluster = cluster

    def include_in_ts(self, ts: 'TestSuite') -> None:
        ts.register_tc(self)

    def success(self, results: str = None) -> None:
        if self.results is None:
            self.results = results
        self.status = TC_PASSED

    def fail(self, results: BaseException) -> None:
        self.status = TC_FAILED
        self.results = results

    def warn(self, results: BaseException) -> None:
        self.status = TC_WARNED
        self.results = results

    def exception(self, results: BaseException) -> None:
        self.status = TC_EXCEPTED
        if isinstance(results, BaseExecutorException):
            self.cluster.log.debug(results)
            self.results = "Remote group exception"
        else:
            if not isinstance(results, KeyboardInterrupt):
                print_exc()
            self.results = results

    def get_summary(self, show_hint: bool = False, show_minimal: bool = False, show_recommended: bool = False) -> str:
        output = ""

        output += " " * (15 - len(self.category))
        output += self.category + "  "

        color = ""
        if self.is_succeeded():
            if self.check_color():
                color = "\x1b[38;5;041m"
                output += " \x1b[48;5;041m\x1b[38;5;232m   OK   \x1b[49m\x1b[39m  "
            else:
                output += "    OK     "
        if self.is_failed():
            if self.check_color():
                color = "\x1b[38;5;196m"
                output += " \x1b[48;5;196m\x1b[38;5;231m  FAIL  \x1b[49m\x1b[39m  "
            else:
                output += "   FAIL    "
        if self.is_warned():
            if self.check_color():
                color = "\x1b[38;5;208m"
                output += " \x1b[48;5;208m\x1b[38;5;231m  WARN  \x1b[49m\x1b[39m  "
            else:
                output += "   WARN    "
        if self.is_excepted():
            if self.check_color():
                color = "\x1b[31m"
                output += " \x1b[41m ERROR? \x1b[49m  "
            else:
                output += "  ERROR?   "

        output += self.id + "  "
        output += self.name + " "

        results = " " + str(self.results)

        output += "." * (146 - len(output) - len(results))
        if self.check_color():
            output += "%s%s\x1b[39m" % (color, results)
        else:
            output += "%s" % (results)

        if show_minimal:
            if self.minimal is None:
                output += ' ' * 15
            else:
                minimal = str(self.minimal)
                output += ' ' * (15-len(minimal)) + minimal

        if show_recommended:
            if self.recommended is None:
                output += ' ' * 14
            else:
                recommended = str(self.recommended)
                output += ' ' * (14-len(recommended)) + recommended

        if (show_hint and isinstance(self.results, (TestFailure, TestWarn)) and self.results.hint is not None):
            output += "\n                  HINT:\n" + textwrap.indent(str(self.results.hint), "                       ")

        return output

    def check_color(self) -> bool:
        for handler in self.cluster.log.handlers:
            if isinstance(handler, log.StdoutHandler) and handler.formatter.colorize:
                return True
        return False

    def get_readable_status(self) -> str:
        if self.is_succeeded():
            return 'ok'
        if self.is_failed():
            return 'fail'
        if self.is_warned():
            return 'warning'
        if self.is_excepted():
            return 'exception'

        return "unknown"

    def is_succeeded(self) -> bool:
        return self.status is TC_PASSED

    def is_failed(self) -> bool:
        return self.status is TC_FAILED

    def is_warned(self) -> bool:
        return self.status is TC_WARNED

    def is_excepted(self) -> bool:
        return self.status is TC_EXCEPTED


class TestCaseNegativeResult(BaseException):

    def __init__(self, message: str, hint: str = None):
        super().__init__(message)
        self.message = message
        self.hint = hint


class TestFailure(TestCaseNegativeResult):
    pass


class TestWarn(TestCaseNegativeResult):
    pass


class TestSuite:

    def __init__(self) -> None:
        self.tcs: List[TestCase] = []

    def register_tc(self, tc: TestCase) -> None:
        self.tcs.append(tc)

    def is_any_test_failed(self) -> bool:
        for tc in self.tcs:
            if tc.is_failed() or tc.is_excepted():
                return True
        return False

    def is_any_test_warned(self) -> bool:
        for tc in self.tcs:
            if tc.is_warned():
                return True
        return False

    def print_final_summary(self, show_minimal: bool = True, show_recommended: bool = True) -> None:
        # pylint: disable-next=line-too-long
        result = "          Group    Status   ID    Test                                                               Actual result"
        if show_minimal:
            result += "        Minimal"
        if show_recommended:
            result += "   Recommended"
        result += "\n"

        colorize = False
        for tc in self.tcs:
            result += "\n" + tc.get_summary(show_minimal=show_minimal, show_recommended=show_recommended)
            colorize = tc.check_color()

        result += "\n\nOVERALL RESULTS: "

        for key, value in sorted(self.get_stats_data().items(), key=lambda _key: badges_weights[_key[0]]):
            colors = ''
            if colorize:
                if key == 'succeeded':
                    colors = "\x1b[48;5;041m\x1b[38;5;232m"
                if key == 'failed':
                    colors = "\x1b[48;5;196m\x1b[38;5;231m"
                if key == 'warned':
                    colors = "\x1b[48;5;208m\x1b[38;5;231m"
                if key == 'excepted':
                    colors = "\x1b[41m"
                result += "%s %s %s \x1b[49m\x1b[39m " % (colors, value ,key.upper())
            else:
                result += "%s %s  " % (value ,key.upper())

        result += "\n"

        print(result)  # pylint: disable=bad-builtin

    def print_final_status(self, logger: log.EnhancedLogger) -> None:
        if self.is_any_test_failed():
            logger.error("\nTEST FAILED"
                      "\nThe environment does not meet the minimal requirements. Check the test report and resolve the issues.")
            return
        if self.is_any_test_warned():
            logger.warning("\nTEST PASSED WITH WARNINGS"
                           "\nThe environment meets the minimal requirements, but is not as recommended. "
                           "Try to check the test report and resolve the issues.")
            return
        logger.info("\nTEST PASSED")

    def get_stats_data(self) -> Dict[str, int]:
        results: Dict[str, int] = {}
        for tc in self.tcs:
            key = 'unknown'
            if tc.is_succeeded():
                key = 'succeeded'
            elif tc.is_failed():
                key = 'failed'
            elif tc.is_warned():
                key = 'warned'
            elif tc.is_excepted():
                key = 'excepted'
            value = results.get(key, 0) + 1
            results[key] = value
        return results

    def save_csv(self, destination_file_path: str, delimiter: str = ';') -> None:
        stream = io.StringIO()

        csv_writer = csv.writer(stream, delimiter=delimiter, quotechar='"', quoting=csv.QUOTE_MINIMAL)
        csv_writer.writerow(['group', 'status', 'test_id', 'test_name', 'current_result', 'minimal_result', 'recommended_result'])
        for tc in self.tcs:
            csv_writer.writerow([
                tc.category.lower(),
                tc.get_readable_status(),
                tc.id,
                tc.name,
                tc.results,
                tc.minimal,
                tc.recommended
            ])

        utils.dump_file({}, stream, destination_file_path, dump_location=False)

    def save_html(self, destination_file_path: str, check_type: str, append_styles: bool = True) -> None:
        stream = io.StringIO()

        stream.write(f'<!DOCTYPE html><html><head><meta charset="utf-8"><title>{check_type} Check Report</title></head>'
                     f'<body><div id="date">{datetime.utcnow()}</div><div id="stats">')
        for key, value in sorted(self.get_stats_data().items(), key=lambda _key: badges_weights[_key[0]]):
            stream.write('<div class="%s">%s %s</div>' % (key, value, key))
        stream.write('</div><h1>%s Check Report</h1><table>' % check_type)
        stream.write('<thead><tr><td>Group</td><td>Status</td><td>ID</td><td>Test</td>'
                     '<td>Actual Result</td><td>Minimal</td><td>Recommended</td></tr></thead><tbody>')
        for tc in self.tcs:
            minimal = ''
            if tc.minimal is not None:
                minimal = str(tc.minimal)
            recommended = ''
            if tc.recommended is not None:
                recommended = str(tc.recommended)
            stream.write('<tr class="%s"><td>%s</td><td><div>%s</div></td>'
                         '<td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td></tr>' %
                         (tc.get_readable_status(),
                          tc.category.lower(),
                          tc.get_readable_status(),
                          tc.id,
                          tc.name,
                          tc.results,
                          minimal,
                          recommended
                          ))
        stream.write('</tbody></table>')
        if append_styles:
            css = utils.read_internal('resources/reports/check_report.css')
            stream.write('<style>\n%s\n</style>' % css)

        stream.write('</body></html>')

        utils.dump_file({}, stream, destination_file_path, dump_location=False)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import functools
from copy import deepcopy
from enum import Flag, auto, IntFlag
from types import FunctionType
from typing import Dict, List, Union, Iterable, Tuple, Optional, Any, Callable, cast, Sequence, Protocol

from ordered_set import OrderedSet

from kubemarine.core import log  # pylint: disable=unused-import
from kubemarine.core import utils, static
from kubemarine.core.connections import ConnectionPool
from kubemarine.core.environment import Environment
from kubemarine.core.errors import KME0006
from kubemarine.core.group import NodeGroup, NodeConfig

_AnyConnectionTypes = Union[str, NodeGroup]


class EnrichmentStage(IntFlag):
    """
    The class acts both as a descriptor of a desirable state of the `KubernetesCluster` object,
    and as a selector of `@enrichment` functions.

    For each stage, the cluster is enriched from the beginning selecting suitable enrichment functions.

    To enrich the cluster at some stage, it must first be enriched at all the previous stages
    unless certain optimizations are applied.
    """

    LIGHT = auto()
    """
    The KubernetesCluster enriched at this stage can only be used to connect to nodes
    to fetch and store data not related to the inventory.
    
    For enrichment function, you rarely need this flag as the only selector,
    but likely need `EnrichmentStage.ALL` instead.
    
    Procedure inventory can be taken into account to enrich connections.
    """

    DEFAULT = auto()
    """
    The KubernetesCluster is enriched at this stage not taking into account the procedure inventory.
    After enrichment, the cluster and its enrichment products should be fully read-only.
    Only enriched inventory can be inspected in this state, but neither context, nor the procedure inventory.
    
    The KubernetesCluster object in this state can be used as backup before running of the `Action`.
    It can also be `evolved` to this state from PROCEDURE state.
    Either evolved (if necessary) or not, it can be used to dump the finalized inventory.
    
    For enrichment function, you never need this flag as the only selector.
    Use `EnrichmentStage.FULL` or other compound selectors.
    """

    PROCEDURE = auto()
    """
    The KubernetesCluster is finally enriched at this stage taking into account the procedure inventory.
    Execution on the cluster must be performed only in ths state.
    
    For enrichment function, you can use this flag as the only selector to only merge the procedure inventory,
    or to only verify the result of the merge.
    In this case, it is necessary to specify procedures for which the function is applicable.
    See also `@enrichment`.
    """

    # Some useful compound selectors of enrichment functions.
    FULL = DEFAULT | PROCEDURE
    ALL = LIGHT | DEFAULT | PROCEDURE
    NONE = 0
    """Special flag typically denoting that nothing is enriched."""

    __str__ = Flag.__str__

    @staticmethod
    def values() -> List['EnrichmentStage']:
        return list(EnrichmentStage)


class _Enrichment(Protocol):
    # The method intentionally lacks of `EnrichmentStage` argument for proper encapsulation
    def __call__(self, __cluster: 'KubernetesCluster') -> Optional[dict]: ...


class EnrichmentFunction(_Enrichment):
    """
    Wrapper around callable that performs enrichment with selectors.
    See `@enrichment`.
    """

    def __init__(self, delegate: _Enrichment, stages: EnrichmentStage, procedures: List[str] = None):
        if EnrichmentStage.DEFAULT in stages:
            if EnrichmentStage.PROCEDURE not in stages:
                raise ValueError("If enrichment function is applied at DEFAULT stage, "
                                 "it should also be applied at PROCEDURE stage")
            if procedures is not None:
                raise ValueError("Enrichment function can be restricted to some procedures "
                                 "only if it is applied at PROCEDURE and/or LIGHT stages")
        elif EnrichmentStage.PROCEDURE in stages and procedures is None:
            raise ValueError("If enrichment function is applied at PROCEDURE stage, but not at DEFAULT stage, "
                             "it should be restricted to some procedures")

        self.delegate = delegate
        self.stages = stages
        self.procedures: Optional[List[str]] = procedures

    # The method intentionally lacks of `EnrichmentStage` argument for proper encapsulation
    def __call__(self, cluster: 'KubernetesCluster') -> Optional[dict]:
        return self.delegate(cluster)

    @property
    def name(self) -> str:
        func = cast(FunctionType, self.delegate)
        return f"{func.__module__}.{func.__qualname__}"


def enrichment(stages: EnrichmentStage, procedures: List[str] = None) -> Callable[[_Enrichment], EnrichmentFunction]:
    """
    Wraps callable that performs enrichment, persisting selectors as the wrapper attributes.
    The selectors are supplied as the arguments of this decorator.

    :param stages: `EnrichmentStage` stages to run this function at.
    :param procedures: list of procedures for which the function is applicable.
    """
    def helper(fn: _Enrichment) -> EnrichmentFunction:
        wrapper = EnrichmentFunction(fn, stages, procedures)
        functools.update_wrapper(wrapper, fn)
        return wrapper

    return helper


@dataclasses.dataclass(repr=False)
class _EnrichmentProducts:
    """States of inventory and context at the particular `EnrichmentStage` of enrichment."""
    inventory: dict
    context: Optional[dict]
    procedure_inventory: Optional[dict]

    nodes_inventory: Dict[str, dict] = dataclasses.field(default_factory=dict)
    nodes: Dict[str, Dict[str, NodeGroup]] = dataclasses.field(default_factory=dict)

    formatted_inventory: Optional[dict] = None
    raw_inventory: Optional[dict] = None


class KubernetesCluster(Environment):

    def __init__(self, inventory: dict, context: dict, procedure_inventory: dict,
                 logger: log.EnhancedLogger,
                 connection_pool: ConnectionPool = None, nodes_context: Dict[str, Any] = None) -> None:
        # Enrichment stage field should not be opened even for read only aims.
        # Such encapsulation ensures that enrichment functions do not know about at what stage they are currently run.
        self._enrichment_stage = EnrichmentStage.NONE
        self._products = _EnrichmentProducts(
            utils.deepcopy_yaml(inventory), deepcopy(context), utils.deepcopy_yaml(procedure_inventory))
        self._previous_products = self._products
        self._products.nodes['nodes'] = self._products.nodes['previous'] = {}

        self._logger = logger

        self._connection_pool: Optional[ConnectionPool] = connection_pool
        self._nodes_context: Optional[Dict[str, Any]] = nodes_context

    def enrich(self, stage: EnrichmentStage,
               *,
               enrichment_fns: List[EnrichmentFunction],
               previous_cluster: Optional['KubernetesCluster']) -> 'KubernetesCluster':
        """
        Enrich the cluster to the state represented by the specified enrichment `stage`.

        :param stage: desirable state of the cluster object.
        :param enrichment_fns: enrichment functions to run.
        :param previous_cluster: cluster enriched at the previous stage.
        :return: this cluster object enriched at the specified stage.
        """
        if stage not in EnrichmentStage.values():
            raise ValueError(f"Target state should be one of ({', '.join(map(str, EnrichmentStage))}), got: {stage}")

        if self._enrichment_stage != EnrichmentStage.NONE:
            raise ValueError("The cluster is already enriched")

        # Flag.name is not None for not compound values in any Python version.
        self.log.verbose(f"Starting {cast(str, stage.name).lower()!r} enrichment")

        self._enrichment_stage = stage
        if stage == EnrichmentStage.DEFAULT:
            self._products.procedure_inventory = None
            args: dict = self.context['execution_arguments']
            args.pop('procedure_config', None)
        if stage == EnrichmentStage.PROCEDURE and previous_cluster is not None:
            self._previous_products = previous_cluster._products  # pylint: disable=protected-access

            # Previous nodes refer to previous cluster. Create new group to surely refer to this cluster.
            self._products.nodes['previous'] = {
                role: self.make_group(group.nodes)
                for role, group in self._previous_products.nodes['nodes'].items()}

        if stage == EnrichmentStage.LIGHT:
            # Need different instance of previous_nodes for LIGHT
            # because we should be able to distinguish initial and final nodes at this stage solely.
            # Still have the same instance of inventory holding all (added & removed) nodes.
            self._products.nodes['previous'] = {}

        # run required fields calculation
        for enrichment_fn in enrichment_fns:
            self.log.verbose(f'Calling fn "{enrichment_fn.name}"')
            inventory = enrichment_fn(self)

            if inventory is not None:
                self._products.inventory = inventory

        # For DEFAULT stage, KubernetesCluster.context should be available during enrichment,
        # but should not be accessed after it to exclude an opportunity of it being dependent on the inventory.
        # This is necessary for proper implementation of KubernetesCluster.evolve().
        if stage == EnrichmentStage.DEFAULT:
            self._products.context = None

        self.log.verbose('Enrichment finished!')

        return self

    @enrichment(EnrichmentStage.ALL)
    def convert_formatted_inventory(self) -> dict:
        products = self._products
        if self._enrichment_stage != EnrichmentStage.LIGHT:
            products.formatted_inventory = self.inventory

        products.raw_inventory = utils.convert_native_yaml(self.inventory)

        if self._enrichment_stage == EnrichmentStage.LIGHT:
            # inventory['nodes'] is already enriched, and the procedure inventory is no longer necessary.
            products.procedure_inventory = None
        elif self._enrichment_stage == EnrichmentStage.PROCEDURE:
            products.procedure_inventory = utils.convert_native_yaml(self.procedure_inventory)

        return utils.deepcopy_yaml(products.raw_inventory)

    @enrichment(EnrichmentStage.LIGHT)
    def init_nodes_context(self) -> None:
        self._connection_pool = self.create_connection_pool(self._get_all_nodes().get_hosts())
        self._nodes_context = {node['connect_to']: {} for node in self.inventory["nodes"]}

    def print_roles_summary(self) -> None:
        ips: Dict[str, OrderedSet[str]] = {}
        for nodes in (self.previous_nodes, self.nodes):
            for role, group in nodes.items():
                if role != 'all':
                    ips.setdefault(role, OrderedSet()).update(group.get_hosts())

        for role, group in (
                ('add_node', self.get_new_nodes()),
                ('remove_node', self.get_nodes_for_removal()),
        ):
            if not group.is_empty():
                ips[role] = OrderedSet(group.get_hosts())

        self.log.debug("Inventory file loaded:")
        for role, hosts in ips.items():
            self.log.debug("  %s %i" % (role, len(hosts)))
            for ip in hosts:
                self.log.debug("    %s" % ip)

    @property
    def previous_nodes(self) -> Dict[str, NodeGroup]:
        """
        Previous nodes of the cluster.
        For the cluster enriched at PROCEDURE stage, this includes nodes to be removed, but does not include added nodes.

        Should be changed only during the enrichment.
        """
        return self._products.nodes['previous']

    @property
    def nodes(self) -> Dict[str, NodeGroup]:
        """
        Final nodes of the cluster.
        For the cluster enriched at PROCEDURE stage, this includes added nodes, but does not include nodes to be removed.

        Should be changed only during the enrichment.
        """
        return self._products.nodes['nodes']

    @property
    def previous_inventory(self) -> dict:
        """
        The previous resulting inventory before procedure inventory is applied.
        For `install` and some other procedures this equals to the resulting `KubernetesCluster.inventory`.

        The property should be read only, and should be examined primarily at PROCEDURE stage or inside the flow.
        """
        return self._previous_products.inventory

    @property
    def inventory(self) -> dict:
        """
        The resulting inventory describing the cluster.

        Should be changed only during the enrichment.
        """
        return self._products.inventory

    @property
    def previous_nodes_inventory(self) -> Dict[str, dict]:
        """
        The previous resulting inventory describing specific nodes before procedure inventory is applied.
        For `install` and some other procedures this equals to the resulting `KubernetesCluster.nodes_inventory`.

        The inventory can be derived from the `KubernetesCluster.previous_inventory` only,
        and should not depend on other sources of information.

        The property should be read only, and should be examined primarily at PROCEDURE stage or inside the flow.
        """
        return self._previous_products.nodes_inventory

    @property
    def nodes_inventory(self) -> Dict[str, dict]:
        """
        The resulting inventory describing specific nodes.

        The inventory can be derived from the main `KubernetesCluster.inventory` only,
        and should not depend on other sources of information.

        Should be changed only during the enrichment.
        """
        return self._products.nodes_inventory

    @property
    def context(self) -> dict:
        """
        Context that stores an intermediate enrichment and execution result.

        At DEFAULT stage, the context is available only during enrichment.
        """
        context = self._products.context
        if context is None:
            raise ValueError("Context is not available after 'default' enrichment stage")

        return context

    @property
    def procedure_inventory(self) -> dict:
        procedure_inventory = self._products.procedure_inventory
        if procedure_inventory is None:
            raise ValueError("Procedure inventory is not available at 'default' or 'light' enrichment stage")

        return procedure_inventory

    @property
    def previous_raw_inventory(self) -> dict:
        """
        Inventory represented in python native objects before applying of any enrichment.

        The property should be read only.
        """
        raw_inventory = self._previous_products.raw_inventory
        if raw_inventory is None:
            raise ValueError("Enrichment is not yet started")

        return raw_inventory

    @property
    def raw_inventory(self) -> dict:
        """
        Inventory represented in python native objects after only procedure inventory is applied,
        but without applying of any other enrichment.

        The property should be read only.
        """
        raw_inventory = self._products.raw_inventory
        if raw_inventory is None:
            raise ValueError("Enrichment is not yet started")

        return raw_inventory

    @property
    def formatted_inventory(self) -> dict:
        """
        Formatted inventory after only procedure inventory is applied,
        but without applying of any other enrichment.

        The property can be examined after 'procedure' enrichment stage
        for maintenance procedures that have specific enrichment functions,
        and that support inventory recreation.

        The property should be read only.
        """
        formatted_inventory = self._products.formatted_inventory
        if formatted_inventory is None:
            raise ValueError("Formatted inventory is not available at 'light' enrichment stage")

        return formatted_inventory

    @property
    def nodes_context(self) -> Dict[str, Any]:
        """Various information about nodes that is not changed during Kubemarine run."""
        if self._nodes_context is None:
            raise ValueError("Nodes' context is available only after 'light' enrichment stage is finished")

        return self._nodes_context

    @property
    def connection_pool(self) -> ConnectionPool:
        if self._connection_pool is None:
            raise ValueError("Connection pool is available only after 'light' enrichment stage is finished")

        return self._connection_pool

    def create_connection_pool(self, hosts: List[str]) -> ConnectionPool:
        nodes = {}
        gateway_nodes = {}
        # iterate over inventory last because it redefines previous inventory
        for inventory in (self.previous_inventory, self.inventory):
            for node in inventory.get('nodes', []):
                nodes[node['connect_to']] = node

            for gateway_node in inventory.get('gateway_nodes', []):
                gateway_nodes[gateway_node['name']] = gateway_node

        return self._create_connection_pool(nodes, gateway_nodes, hosts)

    def _create_connection_pool(self, nodes: Dict[str, dict], gateway_nodes: Dict[str, dict], hosts: List[str]) -> ConnectionPool:
        return ConnectionPool(nodes, gateway_nodes, hosts)

    def _create_cluster_storage(self, context: dict) -> utils.ClusterStorage:
        return utils.ClusterStorage(self, context)

    @property
    def log(self) -> log.EnhancedLogger:
        return self._logger

    @property
    def globals(self) -> dict:
        return static.GLOBALS

    def make_group(self, ips: Iterable[_AnyConnectionTypes]) -> NodeGroup:
        return NodeGroup(ips, self)

    def get_access_address_from_node(self, node: NodeConfig) -> str:
        """
        Returns address which should be used to connect to the node via Fabric.
        The address also can be used as unique identifier of the node.
        """
        address: str
        if node.get('connect_to') is not None:
            address = node['connect_to']
        elif node.get('address') is not None:
            address = node['address']
        else:
            address = node['internal_address']

        return address

    def get_nodes_by_names(self, node_names: List[str]) -> List[NodeConfig]:
        result = []
        for node in self.inventory["nodes"]:
            if node['name'] in node_names:
                result.append(node)

        return result

    def get_node_by_name(self, node_name: str) -> Optional[NodeConfig]:
        nodes = self.get_nodes_by_names([node_name])
        return next(iter(nodes), None)

    def get_addresses_from_node_names(self, node_names: List[str]) -> List[str]:
        return [self.get_access_address_from_node(node)
                for node in self.get_nodes_by_names(node_names)]

    def get_node(self, host: _AnyConnectionTypes) -> NodeConfig:
        return self.make_group([host]).get_config()

    def get_node_name(self, host: _AnyConnectionTypes) -> str:
        name: str = self.get_node(host)['name']
        return name

    def make_group_from_nodes(self, node_names: List[str]) -> NodeGroup:
        ips = self.get_addresses_from_node_names(node_names)
        return self.make_group(ips)

    def make_group_from_roles(self, roles: Sequence[str]) -> NodeGroup:
        return self.nodes['all'].having_roles(roles)

    def get_new_nodes(self) -> NodeGroup:
        return self.nodes['all'].exclude_group(self.previous_nodes['all'])

    def get_new_nodes_or_self(self) -> NodeGroup:
        new_nodes = self.get_new_nodes()
        if not new_nodes.is_empty():
            return new_nodes
        return self.nodes['all']

    def get_nodes_for_removal(self) -> NodeGroup:
        return self.previous_nodes['all'].exclude_group(self.nodes['all'])

    def get_changed_nodes(self) -> NodeGroup:
        return self.get_new_nodes().include_group(self.get_nodes_for_removal())

    def get_unchanged_nodes(self) -> NodeGroup:
        return self._get_all_nodes().exclude_group(self.get_changed_nodes())

    def _get_all_nodes(self) -> NodeGroup:
        """Returns literally all nodes including added or removed"""
        return self.nodes['all'].include_group(self.previous_nodes['all'])

    def create_group_from_groups_nodes_names(self, groups_names: List[str], nodes_names: List[str]) -> NodeGroup:
        common_group = self.make_group_from_roles(groups_names)

        if nodes_names:
            common_group = common_group.include_group(self.make_group_from_nodes(nodes_names))

        return common_group

    def schedule_cumulative_point(self, point_method: Callable) -> None:
        self._check_within_flow()

        func = cast(FunctionType, point_method)
        point_fullname = func.__module__ + '.' + func.__qualname__

        if self.context['execution_arguments'].get('disable_cumulative_points', False):
            self.log.verbose('Method %s not scheduled - cumulative points disabled' % point_fullname)
            return

        if point_fullname in self.context['execution_arguments']['exclude_cumulative_points_methods']:
            self.log.verbose('Method %s not scheduled - it set to be excluded' % point_fullname)
            return

        scheduled_points = self.context.get('scheduled_cumulative_points', [])

        if point_method not in scheduled_points:
            scheduled_points.append(point_method)
            self.context['scheduled_cumulative_points'] = scheduled_points
            self.log.verbose('Method %s scheduled' % point_fullname)
        else:
            self.log.verbose('Method %s already scheduled' % point_fullname)

    def is_task_completed(self, task_path: str) -> bool:
        self._check_within_flow()
        return task_path in self.context['proceeded_tasks']

    def _check_within_flow(self, check: bool = True) -> None:
        if check != ('proceeded_tasks' in self.context):
            raise NotImplementedError(f"The method is called {'not ' if check else ''}within tasks flow execution")

    def check_nodes_accessibility(self, skip_check_iaas: bool = True) -> None:
        """Check nodes access statuses"""

        procedure: str = self.context['initial_procedure']
        if procedure == 'check_iaas' and skip_check_iaas:
            return
        
        if procedure == 'remove_node':
            group = self.make_group_from_roles(['control-plane', 'balancer'])
        else:
            group = self.make_group_from_roles(['control-plane', 'balancer']).include_group(self.get_new_nodes_or_self())

        # Check that nodes are online
        remained_offline = group.get_online_nodes(False)

        # Check that nodes are accessible.
        nodes = group.include_group(self.get_changed_nodes())
        inaccessible_online = nodes.get_online_nodes(True).exclude_group(nodes.get_accessible_nodes())

        if not remained_offline.is_empty() or not inaccessible_online.is_empty():
            raise KME0006(remained_offline.get_hosts(), inaccessible_online.get_hosts())

    def get_os_family_for_node(self, host: str) -> str:
        os_family: Optional[str] = self.nodes_context.get(host, {}).get('os', {}).get('family')
        if os_family is None:
            raise Exception('Node %s do not contain necessary context data' % host)
        return os_family

    def get_os_family_for_nodes(self, hosts: Iterable[str]) -> str:
        """
        Returns the detected operating system family for hosts.
        The method skips inaccessible nodes unless all nodes are inaccessible.

        :return: Detected OS family, possible values: "debian", "rhel", "rhel8", "rhel9",
                 "multiple", "unknown", "unsupported", "<undefined>".
        """
        os_families = {self.get_os_family_for_node(host) for host in hosts}
        if os_families == {'<undefined>'}:
            return '<undefined>'
        os_families.discard('<undefined>')

        if len(os_families) > 1:
            return 'multiple'
        elif len(os_families) == 0:
            raise Exception('Cannot get os family for empty nodes list')
        else:
            return list(os_families)[0]

    def get_os_family(self) -> str:
        """
        Returns common OS family name from all final remote hosts.
        The method skips inaccessible nodes unless all nodes are inaccessible.

        :return: Detected OS family, possible values: "debian", "rhel", "rhel8", "rhel9",
                 "multiple", "unknown", "unsupported", "<undefined>".
        """
        return self.nodes['all'].get_nodes_os()

    def get_os_identifiers(self) -> Dict[str, Tuple[str, str]]:
        """
        For each final and accessible node of the cluster, returns a tuple of OS (family, version).
        """
        os_ids = {}
        for host in self.nodes['all'].get_accessible_nodes().get_hosts():
            os_details = self.nodes_context[host]['os']
            os_ids[host] = (os_details['family'], os_details['version'])

        return os_ids

    def _get_associations(self, os_family: str) -> Dict[str, dict]:
        # Iterate over the resulting inventory first because it has priority.
        # Still need to check previous inventory if it contains OS specific section for nodes to be removed.
        for inventory in (self.inventory, self.previous_inventory):
            associations: Optional[dict] = (inventory.get('services', {}).get('packages', {})
                                            .get('associations', {}).get(os_family))
            if associations is not None:
                return associations

        raise Exception(f"Failed to get associations for {os_family!r} OS family")

    def get_associations(self) -> Dict[str, dict]:
        """
        Returns association for all packages from inventory for the cluster.
        The method can be used only if cluster has nodes with the same and supported OS family.
        """
        return self._get_associations(self.get_os_family())

    def _get_associations_for_os(self, os_family: str, package: str) -> dict:
        associations = self._get_associations(os_family).get(package)
        if associations is None:
            raise Exception(f'Failed to get associations for package "{package}"')

        return associations

    def get_associations_for_node(self, host: str, package: str) -> dict:
        """
        Returns all packages associations for specific node.

        :param host: The address of the node for which required to find the associations
        :param package: The package name to get the associations for
        :return: Dict with packages and their associations
        """
        node_os_family = self.get_os_family_for_node(host)
        return self._get_associations_for_os(node_os_family, package)

    def _get_package_associations_for_os(self, os_family: str, package: str, association_key: str) -> Any:
        associations = self._get_associations_for_os(os_family, package)
        association_value = associations.get(association_key)
        if association_value is None:
            raise Exception(f'Failed to get association "{association_key}" for package "{package}"')
        if not isinstance(association_value, str) and not isinstance(association_value, list):
            raise Exception(f'Unsupported association "{association_key}" value type for package "{package}", '
                            f'got: {str(association_value)}')

        return association_value

    def get_package_association(self, package: str, association_key: str) -> Any:
        """
        Returns the specified association for the specified package from inventory for the cluster.
        The method can be used only if cluster has nodes with the same and supported OS family.

        :param package: The package name to get the association for
        :param association_key: Association key to get
        :return: Association string or list value
        """
        os_family = self.get_os_family()
        return self._get_package_associations_for_os(os_family, package, association_key)

    def get_package_association_for_node(self, host: str, package: str, association_key: str) -> Any:
        """
        Returns the specified association for the specified package from inventory for specific node.

        :param host: The address of the node for which required to find the association
        :param package: The package name to get the association for
        :param association_key: Association key to get
        :return: Association string or list value
        """
        os_family = self.get_os_family_for_node(host)
        return self._get_package_associations_for_os(os_family, package, association_key)

    def evolve(self) -> 'KubernetesCluster':
        """
        If the cluster was enriched at PROCEDURE stage, make the cluster be enriched at DEFAULT stage.

        This allows to perform iterative changes in the inventory by sequentially applying the procedure enrichment.
        :return: this cluster object enriched at DEFAULT stage.
        """
        if self._enrichment_stage not in (EnrichmentStage.DEFAULT, EnrichmentStage.PROCEDURE):
            raise ValueError("Cluster instance can be evolved only if being in DEFAULT or PROCEDURE state")

        self._enrichment_stage = EnrichmentStage.DEFAULT
        self._products.context = None
        self._products.procedure_inventory = None
        self._previous_products = self._products

        return self

    def make_finalized_inventory(self, finalization_functions: List[Callable[['KubernetesCluster', dict], dict]]) \
            -> dict:
        prepared_inventory = utils.deepcopy_yaml(self.inventory)
        for finalize_fn in finalization_functions:
            prepared_inventory = finalize_fn(self, prepared_inventory)

        return prepared_inventory

    def preserve_inventory(self, context: dict, *, enriched: bool = True) -> None:
        self.log.debug("Start preserving of the information about the procedure.")
        cluster_storage = self._create_cluster_storage(context)
        cluster_storage.make_dir()
        if self.context.get('initial_procedure') == 'add_node':
            cluster_storage.upload_info_new_control_planes()
        cluster_storage.collect_procedure_info()
        cluster_storage.compress_archive(enriched)
        cluster_storage.upload_and_rotate()



# Copyright 2021-2023 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import socket
import sys

action = sys.argv[1]
port = int(sys.argv[2])
address = sys.argv[3]
ip_version = sys.argv[4]

proto = '{{ proto }}'

s_type = socket.SOCK_STREAM
if proto == 'udp':
    s_type = socket.SOCK_DGRAM

family = socket.AF_INET
if ip_version == '6':
    family = socket.AF_INET6

s = socket.socket(family, s_type)
try:
    s.settimeout(int('{{ timeout }}'))

    sz = int('{{ mtu }}')

    if proto == 'udp':
        s.sendto(os.urandom(sz), (address, port))
        data, server = s.recvfrom(sz)
    else:
        s.connect((address, port))
        data = bytearray()
        if action == 'send':
            s.sendall(os.urandom(sz))
            while len(data) < sz:
                data.extend(s.recv(sz))
finally:
    s.close()

if action == 'send' and len(data) != sz:
    sys.stdout.write("Data is lost\n")
    sys.stdout.flush()
    sys.exit(1)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser
import io
import re
import socket
import time
from typing import Dict, Tuple, Optional, List, Callable

import paramiko
from dateutil.parser import parse
from ordered_set import OrderedSet

from kubemarine import selinux, apparmor, sysctl, modprobe
from kubemarine.core import utils, static
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.core.executor import RunnersResult, Token, GenericResult, Callback, RawExecutor
from kubemarine.core.group import (
    GenericGroupResult, RunnersGroupResult, GroupResultException,
    NodeGroup, DeferredGroup, AbstractGroup, GROUP_RUN_TYPE, CollectorCallback
)
from kubemarine.core.annotations import restrict_empty_group


@enrichment(EnrichmentStage.FULL)
def verify_inventory(cluster: KubernetesCluster) -> None:

    if cluster.inventory['services']['ntp'].get('chrony', {}).get('servers') \
        and (cluster.inventory['services']['ntp'].get('timesyncd', {}).get('Time', {}).get('NTP') or
             cluster.inventory['services']['ntp'].get('timesyncd', {}).get('Time', {}).get('FallbackNTP')):
        raise Exception('chrony and timesyncd configured both at the same time')


@enrichment(EnrichmentStage.FULL)
def enrich_etc_hosts(cluster: KubernetesCluster) -> None:
    # enrich only etc_hosts_generated object, etc_hosts remains as it is
    inventory = cluster.inventory
    # if by chance cluster.yaml contains non empty etc_hosts_generated we have to reset it
    inventory['services']['etc_hosts_generated'] = {}

    control_plain = inventory['control_plain']['internal']

    control_plain_names = []
    control_plain_names.append(cluster.inventory['cluster_name'])
    control_plain_names.append('control-plain')
    control_plain_names = list(OrderedSet(control_plain_names))
    inventory['services']['etc_hosts_generated'][control_plain] = control_plain_names

    for node in inventory['nodes']:
        internal_node_ip_names: List[str] = inventory['services']['etc_hosts_generated'].get(node['internal_address'], [])

        internal_node_ip_names.append("%s.%s" % (node['name'], cluster.inventory['cluster_name']))
        internal_node_ip_names.append(node['name'])
        internal_node_ip_names = list(OrderedSet(internal_node_ip_names))
        inventory['services']['etc_hosts_generated'][node['internal_address']] = internal_node_ip_names

        if node.get('address'):
            external_node_ip_names: List[str] = inventory['services']['etc_hosts_generated'].get(node['address'], [])

            external_node_ip_names.append("%s-external.%s" % (node['name'], cluster.inventory['cluster_name']))
            external_node_ip_names.append(node['name'] + "-external")
            external_node_ip_names = list(OrderedSet(external_node_ip_names))
            inventory['services']['etc_hosts_generated'][node['address']] = external_node_ip_names


@enrichment(EnrichmentStage.LIGHT)
def detect_nodes_context(cluster: KubernetesCluster) -> None:
    """The method should fetch only node specific information that is not changed during Kubemarine run"""
    logger = cluster.log
    logger.debug('Start detecting nodes context...')

    whoami(cluster)
    logger.verbose('Whoami check finished')

    detect_active_interface(cluster)
    logger.verbose('Interface check finished')
    detect_os_family(cluster)
    logger.verbose('OS family check finished')

    logger.debug('Detecting nodes context finished!')


def fetch_os_versions(cluster: KubernetesCluster) -> RunnersGroupResult:
    '''
    For Red Hat, CentOS, Oracle Linux, and Ubuntu information in /etc/os-release /etc/redhat-release is sufficient but,
    Debian stores the full version in a special file. sed transforms version string, eg 10.10 becomes DEBIAN_VERSION="10.10"  
    '''
    group = cluster.make_group(cluster.nodes_context).get_accessible_nodes()

    return group.run(
        "cat /etc/*elease; cat /etc/debian_version 2> /dev/null | sed 's/\\(.\\+\\)/DEBIAN_VERSION=\"\\1\"/' || true")


def detect_os_family(cluster: KubernetesCluster) -> None:
    results = fetch_os_versions(cluster)

    for host, node_context in cluster.nodes_context.items():
        node_context['os'] = {
            'name': '<undefined>',
            'version': '<undefined>',
            'family': '<undefined>'
        }
        if host not in results:
            continue

        stdout = results[host].stdout.lower()

        version = None
        lines = ''

        version_regex = re.compile("\\s\\d*\\.\\d*", re.M)
        for line in stdout.split("\n"):
            if 'centos' in line or 'rhel' in line or 'rocky' in line:
                # CentOS and Red Hat have a major version in VERSION_ID string
                matches = re.findall(version_regex, line)
                if matches:
                    version = matches[0].strip()
            if '=' in line:
                lines += line + "\n"

        os_release = configparser.ConfigParser()
        os_release.read_string("[system]\n" + lines)
        name = os_release.get("system", "id").replace('"', '')
        if version is None:
            if name == 'debian':
                version = os_release.get("system", "debian_version").replace('"', '')
            else:
                # Oracle Linux and Ubuntu have full version in VERSION_ID string
                version = os_release.get("system", "version_id").replace('"', '')

        cluster.log.debug("Distribution: %s; Version: %s" % (name, version))

        os_family = detect_os_family_by_name_version(name, version)

        cluster.log.debug("OS family: %s" % os_family)

        node_context["os"].update({
            'name': name,
            'version': version,
            'family': os_family
        })


def detect_os_family_by_name_version(name: str, version: str) -> str:
    os_family = 'unsupported'
    if name in static.GLOBALS["compatibility_map"]["distributives"]:
        os_family = 'unknown'
        os_family_list = static.GLOBALS["compatibility_map"]["distributives"][name]
        for os_family_item in os_family_list:
            if version in os_family_item["versions"]:
                os_family = os_family_item["os_family"]
                break

    return os_family


def update_resolv_conf(group: NodeGroup, config: dict) -> None:
    # TODO: Use Jinja template
    buffer = get_resolv_conf_buffer(config)
    utils.dump_file(group.cluster, buffer, 'resolv.conf')
    group.put(buffer, "/etc/resolv.conf", backup=True, immutable=True, sudo=True)


def get_resolv_conf_buffer(config: dict) -> io.StringIO:
    buffer = io.StringIO()
    if config.get("search") is not None:
        buffer.write("search %s\n" % config["search"])
    if config.get("nameservers") is not None:
        for address in config["nameservers"]:
            buffer.write("nameserver %s\n" % address)
    return buffer


def generate_etc_hosts_config(inventory: dict, etc_hosts_part: str = 'etc_hosts_generated') -> str:
# generate records for /etc/hosts from services.etc_hosts or services.etc_hosts_generated

    result = ""

    max_len_ip = 0

    for ip in list(inventory['services'][etc_hosts_part].keys()):
        max_len_ip = max(max_len_ip, len(ip))

    for ip, names in inventory['services'][etc_hosts_part].items():
        if isinstance(names, list):
            # remove records with empty values from list
            names = list(filter(len, names))
            # if list is empty, then skip
            if not names:
                continue
            names = " ".join(names)
        result += "%s%s  %s\n" % (ip, " " * (max_len_ip - len(ip)), names)

    return result


def update_etc_hosts(group: NodeGroup, config: str) -> None:
    utils.dump_file(group.cluster, config, 'etc_hosts')
    group.put(io.StringIO(config), "/etc/hosts", backup=True, sudo=True)


def service_status(group: AbstractGroup[GROUP_RUN_TYPE], name: str, callback: Callback = None) -> GROUP_RUN_TYPE:
    return group.sudo('systemctl status %s' % name, warn=True, callback=callback)


def stop_service(group: AbstractGroup[GROUP_RUN_TYPE], name: str, callback: Callback = None) -> GROUP_RUN_TYPE:
    return group.sudo('systemctl stop %s' % name, callback=callback)


def start_service(group: AbstractGroup[GROUP_RUN_TYPE], name: str, warn: bool = False,
                  callback: Callback = None) -> GROUP_RUN_TYPE:
    return group.sudo('systemctl start %s' % name, warn=warn, callback=callback)


def restart_service(group: AbstractGroup[GROUP_RUN_TYPE], name: str = None,
                    callback: Callback = None) -> GROUP_RUN_TYPE:
    if name is None:
        raise Exception("Service name can't be empty")
    return group.sudo('systemctl restart %s' % name, callback=callback)


def enable_service(group: AbstractGroup[GROUP_RUN_TYPE], name: str = None,
                   now: bool = True, callback: Callback = None) -> GROUP_RUN_TYPE:
    if name is None:
        raise Exception("Service name can't be empty")

    cmd = 'systemctl enable %s' % name
    if now:
        cmd = cmd + " --now"
    return group.sudo(cmd, callback=callback)


def disable_service(group: AbstractGroup[GROUP_RUN_TYPE], name: str = None,
                    now: bool = True, callback: Callback = None) -> GROUP_RUN_TYPE:
    if name is None:
        raise Exception("Service name can't be empty")

    cmd = 'systemctl disable %s' % name
    if now:
        cmd = cmd + " --now"
    return group.sudo(cmd, callback=callback)


def patch_systemd_service(group: DeferredGroup, service_name: str, patch_source: str) -> None:
    group.sudo(f"mkdir -p /etc/systemd/system/{service_name}.service.d")
    group.put(io.StringIO(utils.read_internal(patch_source)),
              f"/etc/systemd/system/{service_name}.service.d/{service_name}.conf",
              sudo=True)
    group.sudo("systemctl daemon-reload")


def configure_sensitive_service(group: NodeGroup, action: Callable[[NodeGroup], bool]) -> bool:
    cluster: KubernetesCluster = group.cluster
    is_updated = group.call(action)
    if is_updated:
        cluster.schedule_cumulative_point(reboot_nodes)
        cluster.schedule_cumulative_point(verify_system)

    return is_updated


def fetch_firewalld_status(group: NodeGroup) -> RunnersGroupResult:
    return group.sudo("systemctl status firewalld", warn=True)


def is_firewalld_disabled(group: NodeGroup) -> Tuple[bool, RunnersGroupResult]:
    result = fetch_firewalld_status(group)
    disabled_status = True

    for node_result in list(result.values()):
        if node_result.return_code != 4 and "disabled" not in node_result.stdout:
            disabled_status = False

    return disabled_status, result


def disable_firewalld(group: NodeGroup) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    already_disabled, result = is_firewalld_disabled(group)

    if already_disabled:
        log.debug("Skipped - FirewallD already disabled or not installed")
        return result

    log.verbose("Trying to stop and disable FirewallD...")

    result = disable_service(group, name='firewalld', now=True)

    cluster.schedule_cumulative_point(reboot_nodes)
    cluster.schedule_cumulative_point(verify_system)

    return result


def is_swap_disabled(group: NodeGroup) -> Tuple[bool, RunnersGroupResult]:
    result = group.sudo("cat /proc/swaps", warn=True)
    disabled_status = True

    for node_result in list(result.values()):
        # is there any other lines excluding first head line?
        if node_result.stdout.strip().split('\n')[1:]:
            disabled_status = False

    return disabled_status, result


def disable_swap(group: NodeGroup) -> Optional[RunnersGroupResult]:
    log = group.cluster.log

    already_disabled, result = is_swap_disabled(group)

    if already_disabled:
        log.debug("Skipped - swap already disabled")
        return result

    log.verbose("Switching swap off...")

    group.sudo('swapoff -a', warn=True)
    group.sudo('sed -i.bak \'/swap/d\' /etc/fstab', warn=True)

    group.cluster.schedule_cumulative_point(reboot_nodes)
    group.cluster.schedule_cumulative_point(verify_system)

    return None


def reboot_nodes(cluster: KubernetesCluster) -> None:
    cluster.get_new_nodes_or_self().call(reboot_group)


def reboot_group(group: NodeGroup, try_graceful: bool = None) -> RunnersGroupResult:
    from kubemarine import kubernetes  # pylint: disable=cyclic-import

    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    if try_graceful is None:
        if 'controlplain_uri' not in cluster.context.keys():
            kubernetes.is_cluster_installed(cluster)

    graceful_reboot = try_graceful is True or \
                      (try_graceful is None and cluster.context['controlplain_uri'] is not None)

    if not graceful_reboot:
        return perform_group_reboot(group)

    log.verbose('Graceful reboot required')

    first_control_plane = cluster.nodes['control-plane'].get_first_member()
    results: Dict[str, RunnersResult] = {}

    for node in group.get_ordered_members_list():
        node_config = node.get_config()
        node_name = node.get_node_name()
        cordon_required = 'control-plane' in node_config['roles'] or 'worker' in node_config['roles']
        if cordon_required:
            res = first_control_plane.sudo(
                kubernetes.prepare_drain_command(cluster, node_name, disable_eviction=False),
                warn=True, pty=True)
            log.verbose(res)
        log.debug(f'Rebooting node "{node_name}"')
        raw_results = perform_group_reboot(node)
        if cordon_required:
            res = first_control_plane.sudo(f'kubectl uncordon {node_name}', warn=True)
            log.verbose(res)
        results.update(raw_results)

    return RunnersGroupResult(cluster, results)


def get_reboot_history(group: NodeGroup) -> RunnersGroupResult:
    return group.sudo('last reboot')


def perform_group_reboot(group: NodeGroup) -> RunnersGroupResult:
    log = group.cluster.log

    initial_boot_history = get_reboot_history(group)
    result = group.sudo(group.cluster.globals['nodes']['boot']['reboot_command'], warn=True)
    log.debug("Waiting for boot up...")
    log.verbose("Initial boot history:\n%s" % initial_boot_history)
    group.wait_for_reboot(initial_boot_history)
    return result


def reload_systemctl(group: AbstractGroup[GROUP_RUN_TYPE]) -> GROUP_RUN_TYPE:
    return group.sudo('systemctl daemon-reload')


def configure_chronyd(group: NodeGroup, retries: int = 60) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log
    chronyd_config = ''

    for server in cluster.inventory['services']['ntp']['chrony']['servers']:
        chronyd_config += "server " + server + "\n"

    if cluster.inventory['services']['ntp']['chrony'].get('makestep'):
        chronyd_config += "\nmakestep " + cluster.inventory['services']['ntp']['chrony']['makestep']

    if cluster.inventory['services']['ntp']['chrony'].get('rtcsync', False):
        chronyd_config += "\nrtcsync"

    utils.dump_file(cluster, chronyd_config, 'chrony.conf')
    group.put(io.StringIO(chronyd_config), '/etc/chrony.conf', backup=True, sudo=True)
    group.sudo('systemctl restart chronyd')
    while retries > 0:
        log.debug("Waiting for time sync, retries left: %s" % retries)
        results = group.sudo('chronyc tracking && sudo chronyc sources')
        if results.stdout_contains("Normal"):
            log.verbose("NTP service reported successful time synchronization, validating...")

            _, _, time_diff = get_nodes_time(group)
            if time_diff > cluster.globals['nodes']['max_time_difference']:
                log.debug("Time is not synced yet")
                log.debug(results)
            else:
                log.debug("Time synced!")
                return results

        else:
            log.debug("Time is not synced yet")
            log.debug(results)
        time.sleep(1)
        retries -= 1

    raise Exception("Time not synced, but timeout is reached")


def configure_timesyncd(group: NodeGroup, retries: int = 120) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log
    timesyncd_config = ''

    for section, options in cluster.inventory['services']['ntp']['timesyncd'].items():
        timesyncd_config += '[%s]' % section
        for option_name, option_value in options.items():
            if isinstance(option_value, list):
                option_value_str = " ".join(option_value)
            else:
                option_value_str = str(option_value)
            timesyncd_config += '\n%s=%s' % (option_name, option_value_str)
        timesyncd_config += '\n\n'

    utils.dump_file(cluster, timesyncd_config, 'timesyncd.conf')
    group.put(io.StringIO(timesyncd_config), '/etc/systemd/timesyncd.conf', backup=True, sudo=True)
    res = group.sudo('timedatectl set-ntp true '
                     '&& sudo systemctl enable --now systemd-timesyncd.service '
                     '&& sudo systemctl restart systemd-timesyncd.service '
                     '&& sudo systemctl status systemd-timesyncd.service')
    log.verbose(res)
    while retries > 0:
        log.debug("Waiting for time sync, retries left: %s" % retries)
        results = group.sudo('timedatectl timesync-status && sudo timedatectl status')
        if results.stdout_contains("synchronized: yes"):
            log.verbose("NTP service reported successful time synchronization, validating...")

            _, _, time_diff = get_nodes_time(group)
            if time_diff > cluster.globals['nodes']['max_time_difference']:
                log.debug("Time is not synced yet")
                log.debug(results)
            else:
                log.debug("Time synced!")
                return results

        else:
            log.debug("Time is not synced yet")
            log.debug(results)
        time.sleep(1)
        retries -= 1

    raise Exception("Time not synced, but timeout is reached")


def verify_system(cluster: KubernetesCluster) -> None:
    group = cluster.get_new_nodes_or_self()
    log = cluster.log
    # this method handles clusters with multiple OS
    os_family = group.get_nodes_os()

    if os_family in ['rhel', 'rhel8', 'rhel9'] and cluster.is_task_completed('prepare.system.setup_selinux'):
        log.debug("Verifying Selinux...")
        selinux_configured, selinux_result, _ = \
            selinux.is_config_valid(group,
                                    state=selinux.get_expected_state(cluster.inventory),
                                    policy=selinux.get_expected_policy(cluster.inventory),
                                    permissive=selinux.get_expected_permissive(cluster.inventory))
        log.debug(selinux_result)
        if not selinux_configured:
            raise Exception("Selinux is still not configured")
    else:
        log.debug('Selinux verification skipped - origin task was not completed')

    if cluster.is_task_completed('prepare.system.setup_apparmor') and os_family == 'debian':
        log.debug("Verifying Apparmor...")
        expected_profiles = cluster.inventory['services']['kernel_security'].get('apparmor', {})
        apparmor_configured = apparmor.is_state_valid(group, expected_profiles)
        if not apparmor_configured:
            raise Exception("Apparmor is still not configured")
    else:
        log.debug('Apparmor verification skipped - origin task was not completed')

    if cluster.is_task_completed('prepare.system.disable_firewalld'):
        log.debug("Verifying FirewallD...")
        firewalld_disabled, firewalld_result = is_firewalld_disabled(group)
        log.debug(firewalld_result)
        if not firewalld_disabled:
            raise Exception("FirewallD is still enabled")
    else:
        log.debug('FirewallD verification skipped - origin disable task was not completed')

    if cluster.is_task_completed('prepare.system.disable_swap'):
        log.debug("Verifying swap...")
        swap_disabled, swap_result = is_swap_disabled(group)
        log.debug(swap_result)
        if not swap_disabled:
            raise Exception("Swap is still enabled")
    else:
        log.debug('Swap verification skipped - origin disable task was not completed')

    if cluster.is_task_completed('prepare.system.modprobe'):
        log.debug("Verifying modprobe...")
        modprobe_valid, _, modprobe_result = modprobe.is_modprobe_valid(group)
        log.debug(modprobe_result)
        if not modprobe_valid:
            raise Exception("Required kernel modules are not presented")
    else:
        log.debug('Modprobe verification skipped - origin setup task was not completed')

    if cluster.is_task_completed('prepare.system.sysctl'):
        verify_sysctl(group)
    else:
        log.debug('Kernel parameters verification skipped - origin setup task was not completed')


def verify_sysctl(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster

    cluster.log.debug("Verifying kernel parameters...")
    sysctl_valid = sysctl.is_valid(group)
    if not sysctl_valid:
        raise Exception("Required kernel parameters are not presented")
    else:
        cluster.log.debug("Required kernel parameters are presented")


def detect_active_interface(cluster: KubernetesCluster) -> None:
    group = cluster.make_group(cluster.nodes_context).get_accessible_nodes()
    collector = CollectorCallback(cluster)
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            detect_interface_by_address(node, node.get_config()['internal_address'], collector=collector)

    for host, node_context in cluster.nodes_context.items():
        interface = '<undefined>'
        if host in collector.result:
            interface = collector.result[host].stdout.rstrip('\n')

        node_context['active_interface'] = interface


def detect_interface_by_address(group: DeferredGroup, address: str, collector: CollectorCallback) -> Token:
    return group.run("/usr/sbin/ip -o a | grep %s | awk '{print $2}'" % address, callback=collector)


def _detect_nodes_access_info(cluster: KubernetesCluster) -> None:
    nodes_context = cluster.nodes_context
    hosts_unknown_status = [host for host, node_context in nodes_context.items() if 'access' not in node_context]
    group_unknown_status = cluster.make_group(hosts_unknown_status)
    if group_unknown_status.is_empty():
        return

    check_active_timeout = int(cluster.globals["nodes"]["remove"]["check_active_timeout"])
    exc = None
    results: GenericGroupResult[GenericResult]
    try:
        # This should invoke sudo last reboot
        results = group_unknown_status.wait_and_get_boot_history(timeout=check_active_timeout)
    except GroupResultException as e:
        exc = e
        results = e.result

    for host, result in results.items():
        access_info = {
            'online': False,
            'accessible': False,
            'sudo': 'No'
        }
        nodes_context[host]['access'] = access_info

        if isinstance(result, BaseException):
            if RawExecutor.is_require_nopasswd_exception(result):
                # The error is thrown only if connection is successful, but something is wrong with sudo access.
                # In general, sudo password is incorrect. In our case, user is not a sudoer, or not a nopasswd sudoer.
                access_info['online'] = True
                access_info['accessible'] = True
            elif isinstance(result, socket.timeout):
                # Usually when node is off. All statuses are unchecked.
                pass
            elif isinstance(result, paramiko.ssh_exception.SSHException):
                # At least, node is on, but something is wrong with ssh credentials (user / identity key)
                access_info['online'] = True
            elif isinstance(result, paramiko.ssh_exception.NoValidConnectionsError):
                # Internal socket error, for example, when ssh daemon is off. All statuses are unchecked.
                pass
            elif isinstance(exc, GroupResultException):
                raise exc
        else:
            access_info['online'] = True
            access_info['accessible'] = True
            access_info['sudo'] = "Yes"


def whoami(cluster: KubernetesCluster) -> RunnersGroupResult:
    '''
    Determines different nodes access information, such as if the node is online, ssh credentials are correct, etc.
    '''
    _detect_nodes_access_info(cluster)

    results = cluster.make_group(cluster.nodes_context).get_sudo_nodes().sudo("whoami")
    for host, result in results.items():
        node_ctx = cluster.nodes_context[host]
        node_ctx['access']['sudo'] = 'Root' if result.stdout.strip() == "root" else 'Yes'
    return results


@restrict_empty_group
def get_nodes_time(group: NodeGroup) -> Tuple[float, Dict[str, float], float]:
    """
    Polls the time from the specified group of nodes, parses it and returns tuple with results.
    :param group: Group of nodes, where timestamps should be detected.
    :return: tuple with max timestamp, dict of parsed timestamps per node and max found time difference.

    Max time required for comparing dates computed on nodes, because we can not detect real time between nodes.
    Max nodes time should be less than minimal compared time from future.

    Timestamp - is a time in milliseconds since epoch.
    """

    # Please, note: this method can not detect time on nodes precisely, since Kubemarine can execute commands not at the
    # same time depending on various factors, for example, restrictions on the number of open sockets.

    parsed_time_per_node: Dict[str, float] = {}

    # TODO: request and parse more accurate timestamp in milliseconds

    raw_results = group.run('date')
    for host, result in raw_results.items():
        parsed_time = parse(result.stdout.strip()).timestamp() * 1000
        parsed_time_per_node[host] = parsed_time

    min_time = min(parsed_time_per_node.values())
    max_time = max(parsed_time_per_node.values())

    return max_time, parsed_time_per_node, max_time-min_time



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
from typing import Union, Optional, List, Dict

from kubemarine.core import utils
from kubemarine.core.executor import RunnersResult, Token, Callback
from kubemarine.core.group import (
    NodeGroup, RunnersGroupResult, AbstractGroup, RunResult, DeferredGroup, GROUP_RUN_TYPE
)

DEBIAN_HEADERS = 'sudo DEBIAN_FRONTEND=noninteractive'


def ls_repofiles(group: NodeGroup) -> RunnersGroupResult:
    return group.sudo('ls -la /etc/apt/sources.list.d')


def backup_repo(group: NodeGroup) -> Optional[RunnersGroupResult]:
    if not group.cluster.inventory['services']['packages']['package_manager']['replace-repositories']:
        group.cluster.log.debug("Skipped - repos replacement disabled in configuration")
        return None
    # all files in directory will be renamed: xxx.repo -> xxx.repo.bak
    # if there already any files with ".bak" extension, they should not be renamed to ".bak.bak"!
    return group.sudo("find /etc/apt/ -type f -name '*.list' | "
                      "sudo xargs -t -iNAME mv -bf NAME NAME.bak")


def add_repo(group: NodeGroup, repo_data: Union[List[str], Dict[str, dict], str]) -> RunnersGroupResult:
    create_repo_file(group, repo_data, get_repo_file_name())
    return group.sudo(f'{DEBIAN_HEADERS} apt clean -q && {DEBIAN_HEADERS} apt update -q', pty=True)


def get_repo_file_name() -> str:
    return '/etc/apt/sources.list.d/predefined.list'


def create_repo_file(group: AbstractGroup[RunResult],
                     repo_data: Union[List[str], Dict[str, dict], str],
                     repo_file: str) -> None:
    # if repo_data is list, then convert it to string using join
    if isinstance(repo_data, list):
        repo_data_str = "\n".join(repo_data) + "\n"
    elif isinstance(repo_data, dict):
        raise Exception("Not supported repositories format for apt package manager")
    else:
        repo_data_str = utils.read_external(repo_data)

    repo_data_stream = io.StringIO(repo_data_str)
    group.put(repo_data_stream, repo_file, sudo=True)


def clean(group: NodeGroup) -> RunnersGroupResult:
    return group.sudo(f"{DEBIAN_HEADERS} apt clean -q", pty=True)


def get_install_cmd(include: Union[str, List[str]], exclude: Union[str, List[str]] = None) -> str:
    if isinstance(include, list):
        include = ' '.join(include)
    command = f'{DEBIAN_HEADERS} apt update -q && {DEBIAN_HEADERS} apt-get install -y -q {include}'

    if exclude is not None:
        raise Exception("Option 'exclude' is not supported for apt package manager")

    # apt fails to install (downgrade) package if it is already present and has higher version,
    # thus we do not need additional checks here (in contrast to yum)
    return command


def install(group: AbstractGroup[GROUP_RUN_TYPE], include: Union[str, List[str]] = None,
            exclude: Union[str, List[str]] = None,
            pty: bool = False, callback: Callback = None) -> GROUP_RUN_TYPE:
    if include is None:
        raise Exception('You must specify included packages to install')

    command = get_install_cmd(include, exclude)

    return group.sudo(command, callback=callback, pty=pty)


def remove(group: AbstractGroup[GROUP_RUN_TYPE], include: Union[str, List[str]] = None, exclude: Union[str, List[str]] = None,
           warn: bool = False, hide: bool = True, pty: bool = False) -> GROUP_RUN_TYPE:
    if include is None:
        raise Exception('You must specify included packages to remove')

    if isinstance(include, list):
        include = ' '.join(include)
    command = f'{DEBIAN_HEADERS} apt-get purge -y -q {include}'

    if exclude is not None:
        raise Exception("Option 'exclude' is not supported for apt package manager")

    return group.sudo(command, warn=warn, hide=hide, pty=pty)


def upgrade(group: AbstractGroup[GROUP_RUN_TYPE], include: Union[str, List[str]] = None,
            exclude: Union[str, List[str]] = None,
            pty: bool = False) -> GROUP_RUN_TYPE:
    if include is None:
        raise Exception('You must specify included packages to upgrade')

    if isinstance(include, list):
        include = ' '.join(include)
    command = f'{DEBIAN_HEADERS} apt update -q && {DEBIAN_HEADERS} apt-get install --only-upgrade -y -q {include}'

    if exclude is not None:
        raise Exception("Option 'exclude' is not supported for apt package manager")

    return group.sudo(command, pty=pty)


def no_changes_found(action: str, result: RunnersResult) -> bool:
    if action not in ('install', 'upgrade', 'remove'):
        raise Exception(f"Unknown action {action}")
    return "0 upgraded, 0 newly installed, 0 to remove" in result.stdout


def search(group: DeferredGroup, package: str, callback: Callback = None) -> Token:
    if package is None:
        raise Exception('You must specify package to search')
    command = f'{DEBIAN_HEADERS} apt show -q {package}'

    return group.sudo(command, pty=True, warn=True, callback=callback)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from kubemarine.core import utils


def reload() -> None:
    GLOBALS.clear()
    GLOBALS.update(_load_globals())

    DEFAULTS.clear()
    DEFAULTS.update(_load_defaults())

    KUBERNETES_VERSIONS.clear()
    KUBERNETES_VERSIONS.update(load_kubernetes_versions())


def load_compatibility_map(filename: str) -> dict:
    return utils.load_yaml(utils.get_internal_resource_path(
        f"resources/configurations/compatibility/internal/{filename}"))


def load_kubernetes_versions() -> dict:
    kubernetes_versions = utils.load_yaml(
        utils.get_internal_resource_path('resources/configurations/compatibility/kubernetes_versions.yaml'))

    return kubernetes_versions


def _load_globals() -> dict:
    globals_ = utils.load_yaml(
        utils.get_internal_resource_path('resources/configurations/globals.yaml'))

    for config_filename in ('kubernetes_images.yaml', 'packages.yaml', 'plugins.yaml', 'thirdparties.yaml'):
        internal_compatibility = load_compatibility_map(config_filename)

        globals_compatibility = globals_['compatibility_map']['software']
        duplicates = set(internal_compatibility) & set(globals_compatibility)
        if duplicates:
            raise Exception(f"Duplicated software {', '.join(repr(s) for s in duplicates)}")

        globals_compatibility.update(internal_compatibility)

    return globals_


def _load_defaults() -> dict:
    return utils.load_yaml(
        utils.get_internal_resource_path('resources/configurations/defaults.yaml'))


GLOBALS: dict = {}
DEFAULTS: dict = {}
KUBERNETES_VERSIONS: dict = {}

reload()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Dict, List

from kubemarine.core import log
from kubemarine.core.group import NodeGroup, RunnersGroupResult


def get_status(group: NodeGroup) -> Dict[str, Dict[str, List[str]]]:
    log = group.cluster.log
    result = group.sudo("apparmor_status --json")
    parsed_result = {}
    for host, node_result in result.items():
        log.verbose('Parsing status for %s...' % host)
        parsed_result[host] = parse_status(log, node_result.stdout)
    print_status(log, parsed_result)
    return parsed_result


def parse_status(logger: log.EnhancedLogger, result_stdout: str) -> Dict[str, List[str]]:
    result: Dict[str, List[str]] = {}
    # Temporary workaround as long as we support OS with AppArmor from 3.0.0 to 3.0.8
    # Ubuntu 22.04.1 has 3.0.4
    # Malformed output is caused by false start `], ` delimiter
    # https://gitlab.com/apparmor/apparmor/-/blob/v3.0.4/binutils/aa_status.c#L537
    # https://gitlab.com/apparmor/apparmor/-/issues/295
    if '"processes": {], ' in result_stdout:
        logger.debug("Patching malformed apparmor_status --json output")
        result_stdout = result_stdout.replace('"processes": {], ', '"processes": {')

    parsed_data = json.loads(result_stdout)
    for profile_name, profile_state in parsed_data['profiles'].items():
        result.setdefault(profile_state, []).append(profile_name)

    return result


def print_status(logger: log.EnhancedLogger, parsed_result: dict) -> None:
    res = "AppArmor Status:"
    for state in parsed_result.keys():
        res += "\n  Profiles in %s mode:" % state
        for profile in parsed_result[state]:
            res += "\n    - %s" % profile
    logger.verbose(res)


def is_state_valid(group: NodeGroup, expected_profiles: Dict[str, List[str]]) -> bool:
    log = group.cluster.log

    log.verbose('Verifying Apparmor modes...')

    parsed_result = get_status(group)
    valid = True

    for host, status in parsed_result.items():
        for state, profiles in expected_profiles.items():
            if not profiles:
                continue
            if state == 'disable':
                for profile in profiles:
                    if any(profile in remote_profiles for remote_profiles in status.values()):
                        valid = False
                        log.verbose('Mode %s is enabled on remote host %s' % (state, host))
                        break
            else:
                if not status.get(state):
                    valid = False
                    log.verbose('Mode %s is not presented on remote host %s' % (state, host))
                    break
                # check if all 'cluster.yaml' settings reflect on particular node
                for profile in profiles:
                    if profile not in status[state]:
                        valid = False
                        log.verbose('Profile %s is not enabled in %s mode on remote host %s' % (profile, state, host))
                        break

    return valid


# TODO: describe what the purpose of that method is
def convert_profile(profile: str) -> str:
    profile = profile.replace('/', '.')
    if profile[0] == '.':
        profile = profile[1:]
    return profile


def configure_apparmor(group: NodeGroup, expected_profiles: dict) -> RunnersGroupResult:
    cmd = ''
    for profile in expected_profiles.get('enforce', []):
        profile = convert_profile(profile)
        cmd += (f'sudo rm -f /etc/apparmor.d/disable/{profile}; '
                f'sudo rm -f /etc/apparmor.d/force-complain/{profile}; ')
    for profile in expected_profiles.get('complain', []):
        profile = convert_profile(profile)
        cmd += (f'sudo rm -f /etc/apparmor.d/disable/{profile}; '
                f'sudo ln -s /etc/apparmor.d/{profile} /etc/apparmor.d/force-complain/; ')
    for profile in expected_profiles.get('disable', []):
        profile = convert_profile(profile)
        cmd += (f'sudo rm -f /etc/apparmor.d/force-complain/{profile}; '
                f'sudo ln -s /etc/apparmor.d/{profile} /etc/apparmor.d/disable/; ')
    cmd += 'sudo systemctl reload apparmor.service && sudo apparmor_status'
    return group.sudo(cmd)


def setup_apparmor(group: NodeGroup) -> bool:
    log = group.cluster.log

    if group.get_nodes_os() != 'debian':
        log.debug("Skipped - Apparmor is supported only on Ubuntu/Debian")
        return False

    expected_profiles = group.cluster.inventory['services']['kernel_security'].get('apparmor', {})
    valid = is_state_valid(group, expected_profiles)

    if valid:
        log.debug("Skipped - Apparmor already correctly configured")
        return False

    group.call(configure_apparmor, expected_profiles=expected_profiles)

    return True



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from kubemarine.core.cluster import KubernetesCluster
from kubemarine.kubernetes.object import KubernetesObject


class StatefulSet(KubernetesObject):

    def __init__(self, cluster: KubernetesCluster, name: str = None, namespace: str = None, obj: dict = None) -> None:
        super().__init__(cluster, kind='StatefulSet', name=name, namespace=namespace, obj=obj)

    def is_actual_and_ready(self) -> bool:
        return self.is_updated() and self.is_ready()

    def is_ready(self) -> bool:
        desired_number = self._obj.get('spec', {}).get('replicas')
        ready_number = self._obj.get('status', {}).get('readyReplicas')
        return desired_number is not None \
            and ready_number is not None \
            and desired_number == ready_number

    def is_updated(self) -> bool:
        desired_number = self._obj.get('spec', {}).get('replicas')
        updated_number = self._obj.get('status', {}).get('updatedReplicas')
        return desired_number is not None \
            and updated_number is not None \
            and desired_number == updated_number

    def is_scheduled(self) -> bool:
        desired_number = self._obj.get('spec', {}).get('replicas')
        current_number = self._obj.get('status', {}).get('currentReplicas')
        return desired_number is not None \
            and current_number is not None \
            and desired_number == current_number



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from kubemarine.core.cluster import KubernetesCluster
from kubemarine.kubernetes.object import KubernetesObject


class Deployment(KubernetesObject):

    def __init__(self, cluster: KubernetesCluster, name: str = None, namespace: str = None, obj: dict = None) -> None:
        super().__init__(cluster, kind='Deployment', name=name, namespace=namespace, obj=obj)

    def is_actual_and_ready(self) -> bool:
        return self.is_ready() and self.is_up_to_date()

    def is_up_to_date(self) -> bool:
        desired_number_scheduled = self._obj.get('spec', {}).get('replicas')
        updated_number_scheduled = self._obj.get('status', {}).get('updatedReplicas')
        return desired_number_scheduled is not None \
            and updated_number_scheduled is not None \
            and desired_number_scheduled == updated_number_scheduled

    def is_ready(self) -> bool:
        desired_number_scheduled = self._obj.get('spec', {}).get('replicas')
        number_ready = self._obj.get('status', {}).get('readyReplicas')
        return desired_number_scheduled is not None \
            and number_ready is not None \
            and desired_number_scheduled == number_ready



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import copy

from typing import Any, OrderedDict, List

from kubemarine import kubernetes, packages, plugins
from kubemarine.core import flow
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.kubernetes import components
from kubemarine.plugins import calico
from kubemarine.procedures import install


def deploy_kubernetes_join(cluster: KubernetesCluster) -> None:

    group = cluster.get_new_nodes().having_roles(['control-plane', 'worker'])

    if group.is_empty():
        # If balancers are added, it is necessary to reconfigure apiServer.certSANs
        write_new_apiserver_certsans(cluster)
        cluster.log.debug("No kubernetes nodes to perform")
        return

    group.having_roles(['control-plane']).call(kubernetes.join_new_control_plane)
    group.having_roles(['worker']).exclude_group(cluster.nodes['control-plane']) \
        .call(kubernetes.init_workers)

    write_new_apiserver_certsans(cluster)

    group.call_batch([
        kubernetes.apply_labels,
        kubernetes.apply_taints
    ])

    cluster.log.debug("Waiting for new kubernetes nodes...")
    kubernetes.wait_for_nodes(group)
    kubernetes.schedule_running_nodes_report(cluster)


def write_new_apiserver_certsans(cluster: KubernetesCluster) -> None:
    # If balancer or control plane is added, apiServer.certSANs are changed.
    # See kubernetes.enrich_inventory()
    new_nodes_require_sans = cluster.get_new_nodes().having_roles(['control-plane', 'balancer'])
    if new_nodes_require_sans.is_empty():
        return

    cluster.log.debug("Write new certificates for kube-apiserver")
    components.reconfigure_components(cluster.nodes['control-plane'], ['kube-apiserver/cert-sans'])


def redeploy_plugins_if_needed(cluster: KubernetesCluster) -> None:
    # redeploy_candidates is a source of plugins that may be redeployed.
    # Some plugins from redeploy_candidates will not be redeployed, because they have "install: false"
    redeploy_candidates = {}
    for plugin, plugin_item in cluster.inventory["plugins"].items():
        if (cluster.previous_inventory["plugins"][plugin] != plugin_item
                # New route reflectors are added with disabled fullmesh
                or (plugin == 'calico' and calico.new_route_reflectors_added(cluster))):
            cluster.log.debug(f"Configuration of {plugin!r} plugin has changed, scheduling it for redeploy")
            redeploy_candidates[plugin] = cluster.inventory["plugins"][plugin]

    plugins.install(cluster, redeploy_candidates)


def cache_installed_packages(cluster: KubernetesCluster) -> None:
    """
    Task which is used to collect already installed packages versions on already existing nodes.
    It is called first during "add_node" procedure,
    so that new nodes install exactly the same packages as on other already existing nodes.
    """
    packages.cache_package_versions(cluster, cluster.inventory, by_initial_nodes=True)


tasks: OrderedDict[str, Any] = collections.OrderedDict(copy.deepcopy(install.tasks))
del tasks["deploy"]["accounts"]
tasks["deploy"]["plugins"] = redeploy_plugins_if_needed
tasks["deploy"]["kubernetes"]["init"] = deploy_kubernetes_join
tasks["cache_packages"] = cache_installed_packages
tasks.move_to_end("cache_packages", last=False)


class AddNodeAction(flow.TasksAction):
    def __init__(self) -> None:
        super().__init__('add node', tasks,
                         cumulative_points=install.cumulative_points, recreate_inventory=True)


def create_context(cli_arguments: List[str] = None) -> dict:

    cli_help = '''
    Script for adding node to Kubernetes cluster.

    How to use:

    '''

    parser = flow.new_procedure_parser(cli_help, tasks=tasks)
    context = flow.create_context(parser, cli_arguments, procedure='add_node')
    return context


def main(cli_arguments: List[str] = None) -> None:
    context = create_context(cli_arguments)
    flow.ActionsFlow([AddNodeAction()]).run_flow(context)


if __name__ == '__main__':
    main()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
from dataclasses import dataclass
from typing import Callable, Optional, List, IO, Tuple, cast, Dict, Union, Mapping

import os
from abc import ABC, abstractmethod

from ordered_set import OrderedSet

from kubemarine import plugins
from kubemarine.core import utils, log
from kubemarine.core.cluster import KubernetesCluster

ERROR_MANIFEST_NOT_FOUND = "Cannot find original {manifest} {path} for {plugin!r} plugin"


@dataclass(frozen=True)
class Identity:
    plugin_name: str
    manifest_id: Optional[str] = None

    @property
    def name(self) -> str:
        return (self.plugin_name
                if self.manifest_id is None
                else f"{self.plugin_name}-{self.manifest_id}")

    def repr_id(self) -> str:
        return f"manifest with ID={self.manifest_id!r}" if self.manifest_id is not None else 'manifest'


def get_default_manifest_path(manifest_identity: Identity, version: str) -> str:
    resource = f"plugins/yaml/{manifest_identity.name}-{version}-original.yaml"
    return utils.get_internal_resource_path(resource)


class Manifest:
    def __init__(self, identity: Identity, stream: IO) -> None:
        self.identity = identity
        self._patched = OrderedSet[str]()
        self._excluded = OrderedSet[str]()
        self._included = OrderedSet[str]()
        self._obj_list = self._load(stream)

    def obj_key(self, obj: dict) -> str:
        return f"{obj['kind']}_{obj['metadata']['name']}"

    def all_obj_keys(self) -> List[str]:
        return [self.obj_key(obj) for obj in self._obj_list]

    def key_index(self, key: str) -> int:
        for i, obj in enumerate(self._obj_list):
            if self.obj_key(obj) == key:
                return i

        raise ValueError(f"{key} not found")

    def has_obj(self, key: str) -> bool:
        return any(self.obj_key(obj) == key for obj in self._obj_list)

    def get_obj(self, key: str, *, patch: bool) -> dict:
        """
        Get manifest object in YAML format for the specified key.
        By default, ensure presence of the object.
        If the object is absent, it may be a sign that we no longer able to fulfill the previous contract.
        Caller methods must explicitly allow absent object when necessary.

        :param key: 'kind' and 'name' of object
        :param patch: boolean value that tells the manifest that the searched object is going to be patched.
        :return: manifest object
        """
        for obj in self._obj_list:
            if self.obj_key(obj) == key:
                if patch:
                    self._patched.add(key)
                return obj

        raise ValueError(f"{key} not found")

    def include(self, index: int, obj: dict) -> None:
        self._obj_list.insert(index, obj)
        key = self.obj_key(obj)
        self._included.add(key)

    def exclude(self, key: str) -> None:
        del self._obj_list[self.key_index(key)]
        self._excluded.add(key)

    @property
    def patched(self) -> List[str]:
        return list(self._patched)

    @property
    def included(self) -> List[str]:
        return list(self._included)

    @property
    def excluded(self) -> List[str]:
        return list(self._excluded)

    def dump(self) -> str:
        """
        The method implements the dumping of the list of objects to the string that includes several YAMLs inside
        """
        with io.StringIO() as stream:
            utils.yaml_structure_preserver().dump_all(self._obj_list, stream)
            result = stream.getvalue()

        return result

    def get_all_container_images(self) -> List[str]:
        images = []
        for key in self.all_obj_keys():
            obj = self.get_obj(key, patch=False)
            for spec_section in ('containers', 'initContainers'):
                containers = (obj.get('spec') or {}).get('template', {}).get('spec', {}).get(spec_section, [])
                for container in containers:
                    image = container['image']
                    if image not in images:
                        images.append(image)

        return images

    def _load(self, stream: IO) -> List[dict]:
        """
        The method implements the parse YAML file that includes several YAMLs inside
        :param stream: stream with manifest content that should be parsed
        :return: list of original objects to enrich in YAML format.
        """
        obj_list = []
        source_yamls = utils.yaml_structure_preserver().load_all(stream)
        yaml_keys = set()
        for source_yaml in source_yamls:
            if source_yaml is None:
                continue
            yaml_key = self.obj_key(source_yaml)
            # check if there is no duplication
            if yaml_key in yaml_keys:
                raise Exception(
                    f"The {yaml_key} object is duplicated, please verify the original yaml")

            yaml_keys.add(yaml_key)
            obj_list.append(source_yaml)

        return obj_list


EnrichmentFunction = Callable[[Manifest], None]


class Processor(ABC):
    def __init__(self, logger: log.VerboseLogger, inventory: dict, manifest_identity: Identity,
                 original_yaml_path: Optional[str], destination_name: Optional[str]) -> None:
        """
        :param logger: VerboseLogger instance
        :param inventory: inventory of the cluster
        :param manifest_identity: A pair of (plugin_name, manifest_id) that uniquely identifies the manifest
        :param original_yaml_path: path to custom manifest
        :param destination_name: custom destination manifest file name
        """
        self.log: log.VerboseLogger = logger
        self.inventory = inventory
        self.manifest_identity = manifest_identity
        self.plugin_name = manifest_identity.plugin_name
        self.manifest_path = self._get_manifest_path(original_yaml_path)
        self.destination_name = self._get_destination(destination_name)

    def get_known_objects(self) -> List[str]:
        """
        :return: list with the 'kind' and 'name' of expected objects in original manifest
        """
        return []

    @abstractmethod
    def get_enrichment_functions(self) -> List[EnrichmentFunction]:
        """
        :return: list of enrichment methods
        """
        pass

    def get_namespace_to_necessary_pss_profiles(self) -> Dict[str, str]:
        """
        :return: Minimal PSS profiles for each namespace to set labels to.
        """
        return {}

    def get_version(self) -> str:
        version: str = self.inventory['plugins'][self.plugin_name]['version']
        return version

    def include(self, manifest: Manifest, index: int, obj: dict) -> None:
        key = manifest.obj_key(obj)
        manifest.include(index, obj)
        self.log.verbose(f"The {key} has been added")

    def exclude(self, manifest: Manifest, key: str) -> None:
        manifest.exclude(key)
        self.log.verbose(f"The {key} has been excluded from result")

    def validate_inventory(self) -> None:
        """
        # Check if original YAML exists
        """
        if not os.path.isfile(self.manifest_path):
            raise Exception(ERROR_MANIFEST_NOT_FOUND.format(
                manifest=self.manifest_identity.repr_id(), path=self.manifest_path, plugin=self.plugin_name))

    def validate_original(self, manifest: Manifest) -> None:
        """
        The method implements default validations for manifest objects

        :param manifest: container to operate with manifest objects
        """
        known_objects = self.get_known_objects()

        # check if there are new objects
        for key in manifest.all_obj_keys():
            if key not in known_objects:
                self.log.verbose(f"The current version of original yaml has a new object: {key}")

        # check if known objects were excluded
        for key in known_objects:
            if not manifest.has_obj(key):
                self.log.verbose(f"The current version of original yaml does not include "
                                 f"the following object: {key}")

    def original_manifest(self) -> Manifest:
        """
        get original YAML and parse it into list of objects
        """
        try:
            with utils.open_utf8(self.manifest_path, 'r') as stream:
                return Manifest(self.manifest_identity, stream)
        except Exception as exc:
            raise Exception(f"Failed to load {self.manifest_identity.repr_id()} from {self.manifest_path} "
                            f"for {self.plugin_name!r} plugin") from exc

    def enrich(self) -> Manifest:
        """
        The method implements full processing for the plugin main manifest.
        """

        manifest = self.original_manifest()
        self.validate_original(manifest)

        # call enrichment functions one by one
        enrichment_functions = self.get_enrichment_functions()
        for fn in enrichment_functions:
            fn(manifest)

        self.log.verbose(f"The total number of patched objects is {len(manifest.patched)} "
                         f"the objects are the following: {manifest.patched}")
        self.log.verbose(f"The total number of added objects is {len(manifest.included)} "
                         f"the objects are the following: {manifest.included}")
        self.log.verbose(f"The total number of excluded objects is {len(manifest.excluded)} "
                         f"the objects are the following: {manifest.excluded}")

        return manifest

    def apply(self, cluster: KubernetesCluster, manifest: Manifest) -> None:
        logger = cluster.log
        enriched_manifest = manifest.dump()
        utils.dump_file(cluster, enriched_manifest, self.destination_name)

        destination = '/etc/kubernetes/%s' % self.destination_name

        # create config for plugin module
        config = {
            "source": io.StringIO(enriched_manifest),
            "destination": destination,
            "do_render": False
        }

        logger.debug(f"Uploading {self.manifest_identity.repr_id()} enriched from {self.manifest_path} "
                     f"for {self.plugin_name!r} plugin...")
        logger.debug("\tDestination: %s" % destination)

        plugins.apply_source(cluster, config)

    def _get_manifest_path(self, custom_manifest_path: Optional[str]) -> str:
        if custom_manifest_path is not None:
            config = {"source": custom_manifest_path}
            manifest_path, _ = plugins.get_source_absolute_pattern(config)
        else:
            manifest_path = get_default_manifest_path(self.manifest_identity, self.get_version())

        return manifest_path

    def _get_destination(self, custom_destination_name: Optional[str]) -> str:
        if custom_destination_name is not None:
            return custom_destination_name

        return f'{self.manifest_identity.name}-{self.get_version()}.yaml'

    def assign_default_pss_labels(self, manifest: Manifest, namespace: str) -> None:
        from kubemarine import admission  # pylint: disable=cyclic-import

        key = f"Namespace_{namespace}"
        if admission.is_security_enabled(self.inventory):
            profile = self.get_namespace_to_necessary_pss_profiles()[namespace]
            target_labels = admission.get_labels_to_ensure_profile(self.inventory, profile)
            if not target_labels:
                return

            source_yaml = manifest.get_obj(key, patch=True)
            labels: dict = source_yaml['metadata'].setdefault('labels', {})
            labels.update(target_labels)
            self.log.verbose(f"The {key} has been patched in 'metadata.labels' with pss labels for {profile!r} profile")

    def find_container_for_patch(self, manifest: Manifest, key: str,
                                 *,
                                 container_name: str, is_init_container: bool) -> Tuple[int, dict]:
        """
        Find container according to the search criteria.

        :param manifest: container to operate with manifest objects
        :param key: 'kind' and 'name' of object
        :param container_name: name of container to assign the image in the spec
        :param is_init_container: whether to search container in 'initContainers' or in 'containers' spec.
        :return: tuple of container index within the spec and the container data.
        """
        pos, container = self._find_optional_container(
            manifest, key,
            container_name=container_name, is_init_container=is_init_container, allow_absent=False)

        return pos, cast(dict, container)

    def _find_optional_container(self, manifest: Manifest, key: str,
                                 container_name: str, is_init_container: bool, allow_absent: bool = False) \
            -> Tuple[int, Optional[dict]]:
        source_yaml = manifest.get_obj(key, patch=True)
        template_spec = source_yaml['spec']['template']['spec']
        spec_containers_section = 'initContainers' if is_init_container else 'containers'
        container_pos, container = next(((i, c) for i, c in enumerate(template_spec[spec_containers_section])
                                         if c['name'] == container_name),
                                        (-1, None))

        if container_pos == -1 and not allow_absent:
            raise ValueError(f"Container {container_name!r} is not found in {spec_containers_section!r} spec of {key}")

        return container_pos, container

    def get_target_image(self, plugin_service: Optional[str] = None, image_key: Optional[str] = 'image') -> str:
        """
        Calculates full image path from the plugin configuration in inventory.

        :param plugin_service: section of plugin that contains the desirable image_key
        :param image_key: property name by which the inventory holds the desirable image
        :return: target image to be used in a container
        """
        plugin_section = self.inventory['plugins'][self.plugin_name]
        registry = plugin_section['installation'].get('registry')
        plugin_service_section = plugin_section
        if plugin_service:
            plugin_service_section = plugin_service_section[plugin_service]
        image: str = plugin_service_section[image_key]
        if registry:
            image = f"{registry}/{image}"

        return image

    def enrich_volume_and_volumemount(self, source_yaml: dict, serviceaccount: str, init_container_name: str = None) -> None:
        """
        Add a secret as volume and mount to the specified container in the manifest.

        :param manifest: The Manifest object to operate with.
        :param key: The key identifying the object in the manifest.
        """

        if 'volumes' not in source_yaml['spec']['template']['spec']:
            source_yaml['spec']['template']['spec']['volumes'] = []

        if 'volumeMounts' not in source_yaml['spec']['template']['spec']['containers'][0]:
            source_yaml['spec']['template']['spec']['containers'][0]['volumeMounts'] = []

        volume_names = [volume["name"] for volume in source_yaml["spec"]["template"]["spec"]["volumes"]]
        if f"{serviceaccount}-token" not in volume_names:
            source_yaml['spec']['template']['spec']['volumes'].append({
                'name': f'{serviceaccount}-token',
                'secret': {
                    'secretName': f'{serviceaccount}-token'  
                }
            })
        source_yaml['spec']['template']['spec']['containers'][0]['volumeMounts'].append({
            'name': f'{serviceaccount}-token',
            'mountPath': '/var/run/secrets/kubernetes.io/serviceaccount'
        })

        init_containers = source_yaml['spec']['template']['spec'].get('initContainers', [])
        target_container = next((c for c in init_containers if c['name'] == init_container_name), None)
        if target_container:
            if 'volumeMounts' not in target_container:
                target_container['volumeMounts'] = []
            if f"{serviceaccount}-token" not in [vm['name'] for vm in target_container['volumeMounts']]:
                target_container['volumeMounts'].append({
                    'name': f'{serviceaccount}-token',
                    'mountPath': '/var/run/secrets/kubernetes.io/serviceaccount'
                })

    def enrich_image_for_container(self, manifest: Manifest, key: str,
                                   *,
                                   plugin_service: Optional[str] = None,
                                   container_name: str, is_init_container: bool,
                                   allow_absent: bool = False) -> None:
        """
        The method patches the image of the specified container.

        :param manifest: container to operate with manifest objects
        :param key: 'kind' and 'name' of object
        :param plugin_service: section of plugin that contains the desirable 'image'
        :param container_name: name of container to assign the image in the spec
        :param is_init_container: whether to search container in 'initContainers' or in 'containers' spec.
        :param allow_absent: if True, and if container is not found, silently do nothing.
        """
        image = self.get_target_image(plugin_service=plugin_service, image_key='image')

        spec_containers_section = 'initContainers' if is_init_container else 'containers'

        container_pos, container = self._find_optional_container(manifest, key,
            container_name=container_name, is_init_container=is_init_container, allow_absent=allow_absent)
        if container is None:
            return

        container['image'] = image
        self.log.verbose(f"The {key} has been patched in "
                         f"'spec.template.spec.{spec_containers_section}.[{container_pos}].image' with {image!r}")

    def enrich_resources_for_container(self, manifest: Manifest, key: str,
                                       *,
                                       plugin_service: Optional[str] = None,
                                       container_name: str) -> None:
        """
        The method patches the resources of the specified container.

        :param manifest: container to operate with manifest objects
        :param key: 'kind' and 'name' of object
        :param plugin_service: section of plugin that contains the desirable 'resources'
        :param container_name: name of container to assign the resources in the spec
        """
        container_pos, container = self.find_container_for_patch(manifest, key,
                                                                 container_name=container_name, is_init_container=False)

        plugin_service_section = self.inventory['plugins'][self.plugin_name]
        if plugin_service:
            plugin_service_section = plugin_service_section[plugin_service]
        container['resources'] = plugin_service_section['resources']

        self.log.verbose(f"The {key} has been patched in "
                         f"'spec.template.spec.containers.[{container_pos}].resources' "
                         f"with {plugin_service_section['resources']!r}")

    def enrich_args_for_container(self, manifest: Manifest, key: str,
                                  *,
                                  plugin_service: str,
                                  container_name: str,
                                  remove_args: List[str] = None,
                                  extra_args: List[str] = None) -> None:
        """
        Add and / or remove the specified arguments to the specified container.

        :param manifest: container to operate with manifest objects
        :param key: 'kind' and 'name' of object
        :param plugin_service: section of plugin that contains the extra 'args'
        :param container_name: name of container to enrich args to
        :param remove_args: list of arguments to remove
        :param extra_args: List of arguments to add.
                           It is currently not possible to override original or added by Kubemarine arguments.
        """
        container_pos, container = self.find_container_for_patch(manifest, key,
                                                                 container_name=container_name, is_init_container=False)
        container_args = container['args']

        if remove_args is None:
            remove_args = []

        for remove_arg in remove_args:
            for i, container_arg in enumerate(container_args):
                if container_arg == remove_arg or container_arg.startswith(remove_arg + "="):
                    del container_args[i]
                    self.log.verbose(f"The {container_arg!r} argument has been removed from "
                                     f"'spec.template.spec.containers.[{container_pos}].args' in the {key}")
                    break
            else:
                self.log.verbose(f"Not found argument {remove_arg!r} to remove from "
                                 f"'spec.template.spec.containers.[{container_pos}].args' in the {key}")

        if extra_args is None:
            extra_args = []

        plugin_service_section = self.inventory['plugins'][self.plugin_name][plugin_service]
        extra_args.extend(plugin_service_section.get('args', []))

        for extra_arg in extra_args:
            extra_arg_key = extra_arg.split('=')[0]
            for i, container_arg in enumerate(container_args):
                if container_arg == extra_arg_key or container_arg.startswith(extra_arg_key + "="):
                    raise Exception(
                        f"{extra_arg_key!r} argument is already defined in "
                        f"'spec.template.spec.containers.[{container_pos}].args' for the {key}.")

            container_args.append(extra_arg)
            self.log.verbose(f"The {extra_arg!r} argument has been added to "
                             f"'spec.template.spec.containers.[{container_pos}].args' in the {key}")

    def enrich_env_for_container(self, manifest: Manifest, key: str,
                                 *,
                                 plugin_service: Optional[str] = None,
                                 container_name: str,
                                 env_delete: List[str] = None,
                                 env_ensure: Mapping[str, Union[str, dict]] = None) -> None:
        """
        Add and / or remove the specified environment variables to the specified container.

        :param manifest: container to operate with manifest objects
        :param key: 'kind' and 'name' of object
        :param plugin_service: section of plugin that contains the extra 'env'
        :param container_name: name of container to enrich env variables to
        :param env_delete: List of env variables to delete.
                           They are deleted irrespective of what is configured in 'env' section of plugin service.
        :param env_ensure: Key-value map of env variables to insert or update.
                           These variables and not configurable.
        :return:
        """
        container_pos, container = self.find_container_for_patch(manifest, key,
                                                                 container_name=container_name, is_init_container=False)

        if env_delete is None:
            env_delete = []

        for name in env_delete:
            for i, e in enumerate(container['env']):
                if e['name'] == name:
                    del container['env'][i]
                    self.log.verbose(f"The {name!r} env variable has been removed from "
                                     f"'spec.template.spec.containers.[{container_pos}].env' in the {key}")
                    break

        env_update: Dict[str, dict] = {}

        def update_env(name: str, value: Union[str, dict]) -> None:
            if isinstance(value, str):
                env_update[name] = {'value': value}
            elif isinstance(value, dict):
                env_update[name] = {'valueFrom': value}
            self.log.verbose(f"The {key} has been patched in "
                             f"'spec.template.spec.containers.[{container_pos}].env.{name}' with '{value}'")

        if env_ensure is None:
            env_ensure = {}

        for name, value in env_ensure.items():
            update_env(name, value)

        plugin_service_section = self.inventory['plugins'][self.plugin_name]
        if plugin_service:
            plugin_service_section = plugin_service_section[plugin_service]

        for name, value in plugin_service_section.get('env', {}).items():
            if name in env_delete:
                continue
            if name in env_ensure:
                raise Exception(f"Environment variable {name} is currently not configurable.")
            update_env(name, value)

        for env in container['env']:
            name = env['name']
            if name not in env_update:
                continue

            value = env_update.pop(name)
            keys = list(env.keys())
            for k in keys:
                if k != 'name' and k not in value:
                    del env[k]

            env.update(value)

        for name, env in env_update.items():
            new_env = {'name' : name}
            new_env.update(env)
            container['env'].append(new_env)

    def enrich_node_selector(self, manifest: Manifest, key: str,
                             *,
                             plugin_service: str) -> None:
        source_yaml = manifest.get_obj(key, patch=True)
        node_selector = self.inventory['plugins'][self.plugin_name][plugin_service]['nodeSelector']
        source_yaml['spec']['template']['spec']['nodeSelector'] = node_selector
        self.log.verbose(f"The {key} has been patched in 'spec.template.spec.nodeSelector' with {node_selector!r}")

    def enrich_tolerations(self, manifest: Manifest, key: str,
                           *,
                           plugin_service: Optional[str] = None,
                           extra_tolerations: List[dict] = None,
                           override: bool = False) -> None:
        source_yaml = manifest.get_obj(key, patch=True)
        template_spec: dict = source_yaml['spec']['template']['spec']
        if override and template_spec.get('tolerations', []):
            del template_spec['tolerations']
            self.log.verbose(f"The 'tolerations' property has been removed from 'spec.template.spec' in the {key}")

        tolerations: List[dict] = []
        if extra_tolerations:
            tolerations.extend(extra_tolerations)
        plugin_service_section = self.inventory['plugins'][self.plugin_name]
        if plugin_service:
            plugin_service_section = plugin_service_section[plugin_service]
        tolerations.extend(plugin_service_section.get('tolerations', []))

        for val in tolerations:
            template_spec.setdefault('tolerations', []).append(val)
            self.log.verbose(f"The {key} has been patched in 'spec.template.spec.tolerations' with '{val}'")


PROCESSOR_PROVIDER = Callable[[log.VerboseLogger, dict, Optional[str], Optional[str]], Processor]
